"""
A module that handles probability calculations according different pscales that
may represent probability coefficients (positive float scalars) or log 
probability offsets (complex float scalars).
"""

#-------------------------------------------------------------------------------
import numpy as np
from probayes.constants import NEARLY_POSITIVE_ZERO, \
                               NEARLY_POSITIVE_INF, \
                               NEARLY_NEGATIVE_INF, \
                               LOG_NEARLY_POSITIVE_INF, \
                               COMPLEX_ZERO

#-------------------------------------------------------------------------------
def iscomplex(pscale):
  """ Returns whether pscale is an instance of a complex number """
  return isinstance(pscale, complex)

#-------------------------------------------------------------------------------
def eval_pscale(pscale=None):
  """ Returns a float or complex pscale with the following conventions:
  if pscale is None, returns 1.
  if pscale is 'log' or 'ln' or 0, returns 0.j.
  otherwise pscale must be real or complex, then it is returned
  """
  if pscale is None:
    return 1.
  if pscale == 1:
    return 1.
  if pscale in ['log', 'ln', 0]:
    return COMPLEX_ZERO
  if isinstance(pscale, int):
    return float(pscale)
  if isinstance(pscale, float):
    if pscale == 0.:
      return COMPLEX_ZERO
    return pscale
  if iscomplex(pscale):
    return pscale
  raise ValueError("Cannot evaluate pscale={}".format(pscale))

#-------------------------------------------------------------------------------
def log_prob(prob):
  """ Safely returns the logarithm of probability values in prob """
  if np.isscalar(prob):
    if prob >= NEARLY_POSITIVE_ZERO:
      return np.log(prob)
    return NEARLY_NEGATIVE_INF
  logp = np.tile(NEARLY_NEGATIVE_INF, prob.shape)
  ok = prob >= NEARLY_POSITIVE_ZERO
  logp[ok] = np.log(prob[ok])
  return logp

#-------------------------------------------------------------------------------
def exp_logp(logp):
  """ Safely returns the exponentional of log probability values in logp """
  if np.isscalar(logp):
    if logp <= LOG_NEARLY_POSITIVE_INF:
      return np.exp(logp)
    return NEARLY_POSITIVE_INF
  prob = np.tile(NEARLY_POSITIVE_INF, logp.shape)
  ok = logp <= LOG_NEARLY_POSITIVE_INF
  prob[ok] = np.exp(logp[ok])
  return prob

#-------------------------------------------------------------------------------
def logp_offs(pscale=None):
  """ Returns the offset in log probability represented by pscale """
  pscale = eval_pscale(pscale)
  if not iscomplex(pscale):
    return float(np.log(pscale))
  if np.abs(np.imag(pscale)) < NEARLY_POSITIVE_ZERO:
    return float(np.real(pscale))
  return -float(np.real(pscale))

#-------------------------------------------------------------------------------
def prob_coef(pscale=None):
  """ Returns the coefficient in probability represented by pscale """
  pscale = eval_pscale(pscale)
  if not iscomplex(pscale):
    return float(pscale)
  if np.abs(np.imag(pscale)) < NEARLY_POSITIVE_ZERO:
    return np.exp(float(np.real(pscale)))
  return np.exp(-float(np.real(pscale)))

#-------------------------------------------------------------------------------
def real_sqrt(real):
  """ Safely returns the square root of positive real numbers """
  if np.isscalar(real):
    if real >= NEARLY_POSITIVE_ZERO:
      return np.sqrt(real)
    return 0.
  root = np.zeros_like(real)
  ok = real >= NEARLY_POSITIVE_ZERO
  root[ok] = np.sqrt(real[ok])
  return root

#-------------------------------------------------------------------------------
def rescale(prob, *args):
  """ Rescales prob according to pscales given in args """
  pscale, rtype = None, None
  prob = prob if np.isscalar(prob) \
           or prob.dtype in [float, np.dtype('float32'), np.dtype('float64')] \
         else np.array(prob, dtype=float)
  if len(args) == 0: 
    return prob
  elif len(args) == 1: 
    rtype = args[0]
  else: 
    pscale, rtype = args[0], args[1]
  pscale, rtype = eval_pscale(pscale), eval_pscale(rtype)
  p_log, r_log = iscomplex(pscale), iscomplex(rtype)
  if p_log == r_log and pscale == rtype:
    return prob

  # Support non-logarithmic conversion (maybe used to avoid logging zeros)
  if not p_log and not r_log:
    coef = pscale / rtype
    if coef == 1.:
      return prob
    else:
      return coef * prob

  # For floating point precision, perform other operations in log-space
  if not p_log: prob = log_prob(prob)
  d_offs = logp_offs(pscale) - logp_offs(rtype)
  if np.abs(d_offs) >= NEARLY_POSITIVE_ZERO: prob = prob + d_offs
  if r_log:
    return prob
  return exp_logp(prob)

#-------------------------------------------------------------------------------
def prod_pscale(pscales, use_logp=None):
  """ Returns the natural product pscale according all the values in pscales,
  adopting a log scale if use_logp is True (which defaults to True if any
  values in pscales are complex.
  """
  if not len(pscales):
    return None
  if use_logp is None:
    use_logp = any([iscomplex(pscale) for pscale in pscales])
  rtype = 0. if use_logp else 1.
  for _pscale in pscales:
    pscale = eval_pscale(_pscale)
    if use_logp:
      rtype += logp_offs(pscale)
    else:
      rtype *= prob_coef(pscale)
  if use_logp:
    if abs(rtype) < NEARLY_POSITIVE_ZERO:
      return COMPLEX_ZERO
    elif rtype > 0:
      return complex(np.log(rtype), 0.)
    else:
      return complex(np.log(-rtype), np.pi)
  return rtype

#-------------------------------------------------------------------------------
def prod_rule(*args, **kwds):
  """ Applies product rule to all probability arrays in arrays with optional
  keywords:

  'pscales': a list of corresponding pscales (defaulting to pscales all 1)
  'use_logp': a boolean flag to return log probabilities (defaults to true if
              any value in pscales is complex)

  :return: joint_probility, pscale (two element tuple)
  """
  kwds = dict(kwds)
  pscales = kwds.get('pscales', [1.] * len(args))
  use_logp = kwds.get('use_logp', any([iscomplex(_pscale) for _pscale in pscales]))
  ppscale = prod_pscale(pscales, use_logp)
  pscale = kwds.get('pscale', ppscale)
  n_args = len(args)
  assert len(pscales) == n_args, \
      "Input pscales length {} incommensurate with number of arguments {}".\
      format(len(pscales), n_args)
  
  def _apply_prod(probs):
    # Numpy sum() and prod() produce inconsistent results with lists
    if len(probs) == 1:
      prob = np.copy(probs[0])
    else:
      prob = probs[0] + probs[1] if use_logp else probs[0] * probs[1]
      for _prob in probs[2:]:
        if use_logp:
          prob = prob + _prob
        else:
          prob = prob * _prob
    return prob

  # Possibly fast-track
  if use_logp != iscomplex(ppscale):
    ppscale = complex(np.log(ppscale), 0.) if use_logp else float(np.exp(ppscale))
  elif use_logp == iscomplex(pscale) and ppscale == pscale and \
      len(set([iscomplex(_pscale) for _pscale in pscales])) == 1:
    prob = _apply_prod(list(args))
    return prob, pscale

  # Otherwise exp/log before evaluating product
  probs = [None] * n_args
  for i, arg in enumerate(args):
    p_log = iscomplex(pscales[i])
    probs[i] = args[i]
    if use_logp:
      if not p_log:
        probs[i] = log_prob(probs[i])
    else:
      if p_log:
        probs[i] = exp_logp(probs[i])
  prob = _apply_prod(probs)
  if use_logp != iscomplex(pscale):
    prob = rescale(prob, ppscale, pscales)

  return prob, pscale

#-------------------------------------------------------------------------------
def div_prob(dividend, divisor, *args, pscale=None):
  """ Safely divides two probability arrays.

  :param dividend: numerator
  :param divisor: denominator
  :param *args: pscales for divident, divisor (defaults to 1. in each case).
  :param pscale: output pscale (defaults to result first pscale).
  """
  pscales = [None, None]
  if len(args):
    assert len(args) == 2, "Both pscales must be specified if at all"
    pscales[0] = eval_pscale(args[0])
    pscales[1] = eval_pscale(args[1])
    pscale = pscale or pscales[0]
  dividend = rescale(dividend, pscales[0], None)
  divisor = rescale(divisor, pscales[1], None)
  quotient = dividend / np.maximum(NEARLY_POSITIVE_ZERO, divisor)
  return rescale(quotient, None, pscale)
  
#-------------------------------------------------------------------------------



""" Provides probability distribution funciontlaity based on Distribution """

#-------------------------------------------------------------------------------
import collections
import numpy as np
from probayes.named_dict import NamedDict
from probayes.pd_utils import str_margcond, margcond_str, product, summate, \
                              rekey_dict, ismonotonic
from probayes.vtypes import isscalar
from probayes.pscales import eval_pscale, rescale, iscomplex
from probayes.pscales import div_prob
from probayes.distribution import Distribution

#-------------------------------------------------------------------------------
class PD (Distribution):
  """ A probability distribution is a distribution with corresponding 
  probabilities. The dimensions of the probability scalar or array must be
  commensurate with the values of the distribution according to their assigned
  dimensions. 

  While it is intended for PD instances to come from RV, RF, SD, or SP calls,
  PDs can be instantiated directly.
  """
  # Protected
  _prob = None    # Probability
  _pscale = None  # Same convention as Prob()
  _marg = None    # Ordered dictionary of marginals: {key: name}
  _cond = None    # Ordered dictionary of conditionals: key: name}

#-------------------------------------------------------------------------------
  def __init__(self, name, *args, **kwds):
    """ Initialises the PD with a name, args, and kwds in the same way as 
    Distribution(), except with the following reserved keywords:
      
      'dims': sets the dimensionality. 
      'prob': sets the probability scalar or array. 
      'pscale': sets the pscale
    """
    args = tuple(args)
    kwds = dict(kwds)
    prob = None if 'prob' not in kwds else kwds.pop('prob')
    pscale = None if 'pscale' not in kwds else kwds.pop('pscale')
    super().__init__(name, *args, **kwds)
    self.pscale = pscale
    self.prob = prob

#-------------------------------------------------------------------------------
  @property
  def name(self):
    return self._name

  @property
  def marg(self):
    return self._marg

  @property
  def cond(self):
    return self._cond

  @name.setter
  def name(self, name):
    # Only the name is sensitive to what are marginal and conditional variables
    NamedDict.name.fset(self, name)
    self._marg, self._cond = str_margcond(self.name)
    
  @property
  def short_name(self):
    marg_keys = list(self._marg.keys())
    cond_keys = list(self._cond.keys())
    marg_str = ','.join(marg_keys)
    cond_str = ','.join(cond_keys)
    return '|'.join([marg_str, cond_str]) if cond_str else marg_str

#-------------------------------------------------------------------------------
  @property
  def pscale(self):
    return self._pscale

  @pscale.setter
  def pscale(self, pscale=None):
    self._pscale = eval_pscale(pscale)

#-------------------------------------------------------------------------------
  @property
  def prob(self):
    return self._prob

  @prob.setter
  def prob(self, prob=None):
    self._prob = prob
    if self._prob is None:
      return
    if self._issingleton:
      assert isscalar(self._prob), "Singleton vals with non-scalar prob"
    else:
      assert not isscalar(self._prob), "Non singleton values with scalar prob"
      assert self._ndim == self._prob.ndim, \
        "Mismatch in dimensionality between values {} and probabilities {}".\
        format(self.ndim, self._prob.ndim)
      assert np.all(np.array(self._shape) == np.array(self._prob.shape)), \
        "Mismatch in dimensions between values {} and probabilities {}".\
        format(self._shape, self._prob.shape)

#-------------------------------------------------------------------------------
  @property
  def dims(self):
    return self._dims

  @dims.setter
  def dims(self, dims=None):
    """ Sets the dimensions for each of the variables.

    :param dims: a dictionary of {variable_name: variable_dim}

    The keys should correspond to that of a dictionary. If dims
    is None, then the dimensionality is set according to the
    order in values.
    """
    Distribution.dims.fset(self, dims)

    # Override name entries for scalar values
    for i, key in enumerate(self._keylist):
      assert key in self._keyset, \
          "Value key {} not found among name keys {}".format(key, self._keyset)
      if self._aresingleton[i]:
        if key in self.marg.keys():
          self.marg[key] = "{}={}".format(key, self[key])
        elif key in self.cond.keys():
          self.cond[key] = "{}={}".format(key, self[key])
        else:
          raise ValueError("Variable {} not accounted for in name {}".format(
                            key, self.name))
    self._name = margcond_str(self.marg, self.cond)

#-------------------------------------------------------------------------------
  def marginalise(self, keys):
    # from p(A, key | B), returns P(A | B)
    if isinstance(keys, str):
      keys = [keys]
    for key in keys:
      assert key in self._marg.keys(), \
        "Key {} not marginal in distribution {}".format(key, self._name)
    keys  = set(keys)
    marg = collections.OrderedDict(self._marg)
    cond = collections.OrderedDict(self._cond)
    vals = collections.OrderedDict()
    dims = collections.OrderedDict()
    dim_delta = 0
    sum_axes = set()
    for i, key in enumerate(self._keylist):
      if key in keys:
        assert not self._aresingleton[i], \
            "Cannot marginalise along scalar for key {}".format(key)
        sum_axes.add(self._dims[key])
        marg.pop(key)
        dim_delta += 1
      else:
        if not self._aresingleton[i]:
          dims.update({key: self._dims[key] - dim_delta})
        vals.update({key:self[key]})
    name = margcond_str(marg, cond)
    prob = rescale(self._prob, self._pscale, 1.)
    sum_prob = np.sum(prob, axis=tuple(sum_axes), keepdims=False)
    prob = rescale(sum_prob, 1., self._pscale)
    return PD(name, vals, dims=dims, prob=prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def marginal(self, keys):
    # from p(A, key | B), returns P(key | B)
    if isinstance(keys, str):
      keys = [keys]

    # Check keys arg marginal
    keys = set(keys)
    dims = set()
    for key in keys:
      assert key in self._marg.keys(), \
        "Key {} not marginal in distribution {}".format(key, self._name)
      dim = self._dims[key]
      if dim is not None:
        dims.add(dim)

    # Check consistency of marginal dims
    for key in self._keylist:
      dim = self._dims[key]
      if dim in dims:
        assert key in keys, \
            "Dimensionality precludes marginalising {} without: {}".\
            format(keys, key)

    # Determine keys to marginalise by exclusion
    marginalise_keys = set()
    aresingletons = []
    marg_scalars = set()
    for i, key in enumerate(self._keylist):
      singleton = self._aresingleton[i]
      marginal = key in keys
      if key in self._marg.keys():
        aresingletons.append(singleton)
        if singleton:
          marg_scalars.add(key)
        if not singleton and not marginal:
          marginalise_keys.add(key)

    # If including any marginal scalars, must include all scalars
    if any(aresingletons):
      assert marg_scalars.issubset(keys), \
        "If evaluating marginal for key {}".format(key) + ", " + \
        "must include all marginal scalars in {}".format(self._marg.keys())

    return self.marginalise(marginalise_keys)
        
#-------------------------------------------------------------------------------
  def conditionalise(self, keys):
    # from P(A, key | B), returns P(A | B, key).
    # if vals[key] is a scalar, this effectively normalises prob
    if isinstance(keys, str):
      keys = [keys]
    keys = set(keys)
    for key in keys:
      assert key in self._marg.keys(), \
        "Key {} not marginal in distribution {}".format(key, self.name)
    dims = collections.OrderedDict()
    marg = collections.OrderedDict(self._marg)
    cond = collections.OrderedDict(self._cond)
    normalise = False
    delta = 0
    marg_scalars = set()
    for i, key in enumerate(self._keylist):
      if key in keys:
        cond.update({key: marg.pop(key)})
      if self._aresingleton[i]:
        dims.update({key: None})
        if key in keys:
          normalise = True
      elif key in self._marg.keys():
        if self._aresingleton[i]:
          marg_scalars.add(key)
        if key in keys:
          delta += 1 # Don't add to dim just yet
        else:
          dim = self._dims[key]
          dims.update({key: dim})
      else:
        dim = self.dims[key] - delta
        dims.update({key: dim})

    # Reduce remaining marginals to lowest dimension
    dim_val = [val for val in dims.values() if val is not None]
    dim_max = 0
    if len(dim_val):
      dim_min = min(dim_val)
      for key in dims.keys():
        if dims[key] is not None:
          dim = dims[key]-dim_min
          dims.update({key: dim})
          dim_max = max(dim_max, dim)
    dim_min = self.ndim
    for key in keys:
      dim = self.dims[key]
      if dim is not None:
        dim_min = min(dim_min, dim)
    for key in keys:
      dim = self._dims[key]
      if dim is not None:
        dims.update({key: dim-dim_min+dim_max+1})
    if normalise:
      assert marg_scalars.issubset(set(keys)), \
        "If conditionalising for key {}".format(key) + "," + \
        "must include all marginal scalars in {}".format(self._marg.keys())

    # Setup vals dimensions and evaluate probabilities
    name = margcond_str(marg, cond)
    vals = collections.OrderedDict(super().redim(dims))
    old_dims = []
    new_dims = []
    sum_axes = set()
    for key in self._keylist:
      old_dim = self._dims[key]
      if old_dim is not None and old_dim not in old_dims:
        old_dims.append(old_dim)
        new_dims.append(dims[key])
        if key not in keys and key in self._marg.keys():
          sum_axes.add(dims[key])
    prob = np.moveaxis(self._prob, old_dims, new_dims)
    if normalise and iscomplex(self._pscale): 
      prob = prob - prob.max()
    prob = rescale(prob, self._pscale, 1.)
    if normalise:
      prob = div_prob(prob, np.sum(prob))
    if len(sum_axes):
      prob = div_prob(prob, \
                         np.sum(prob, axis=tuple(sum_axes), keepdims=True))
    prob = rescale(prob, 1., self._pscale)
    return PD(name, vals, dims=dims, prob=prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def redim(self, dims):
    """ 
    Returns a distribution according to redimensionised values in dims, index-
    ordered by the order in dims
    """
    dist = super().redim(dims)
    vals, dims = dict(dist), dist.dims
    prob = self._prob

    # Need to realign prob axes to new dimensions
    if not self._issingleton:
      old_dims = []
      new_dims = []
      for i, key in enumerate(self._keylist):
        if not self._aresingletons[i]:
          old_dims.append(self._dims[key])
          new_dims.append(dims[key])
      prob = np.moveaxis(prob, old_dims, new_dims)

    return PD(self._name, vals, dims=dims, prob=prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def rekey(self, keymap):
    """
    Returns a distribution with modified key names without axes changes.
    """
    dist = super().rekey(keymap)
    marg = rekey_dict(self._marg, keymap) 
    cond = rekey_dict(self._cond, keymap)
    name = margcond_str(marg, cond)
    return PD(name, dict(dist), dims=dist.dims, 
              prob=self.prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def prod(self, keys):
    # from P(A, key | B), returns P(A, {} | B)
    if isinstance(keys, str):
      keys = [keys]
    for key in keys:
      assert key in self.marg.keys(), \
        "Key {} not marginal in distribution {}".format(key, self.name)
    keys  = set(keys)
    marg = collections.OrderedDict(self._marg)
    cond = collections.OrderedDict(self._cond)
    vals = collections.OrderedDict()
    dims = collections.OrderedDict()
    dim_delta = 0
    prod_axes = []
    for i, key in enumerate(self._keylist):
      if key in keys:
        assert not self._aresingleton[i], \
            "Cannot apply product along scalar for key {}".format(key)
        if self._dims[key] not in prod_axes:
          prod_axes.append(self.dims[key])
          dim_delta += 1
        marg.update({key: key+"={}"})
        vals.update({key: {self[key].size}})
      else:
        if not self._aresingleton[i]:
          dims.update({key: self._dims[key] - dim_delta})
        vals.update({key:self[key]})
    name = margcond_str(marg, cond)
    pscale = self._pscale
    pscale_product = pscale
    if pscale_product not in [0., 1.]:
      pscale_scaling = np.prod(np.array(self.shape)[prod_axes])
      if iscomplex(pscale):
        pscale_product += pscale*pscale_scaling 
      else:
        pscale_product *= pscale**pscale_scaling 
    prob = np.sum(self.prob, axis=tuple(prod_axes)) if iscomplex(pscale) \
           else np.prod(self.prob, axis=tuple(prod_axes))
    return PD(name, vals, dims=dims, prob=prob, pscale=pscale_product)

#-------------------------------------------------------------------------------
  def expectation(self, keys=None, exponent=None):
    keys = keys or self.marg.keys()
    if isinstance(keys, str):
      keys = [keys]
    for key in keys:
      assert key in self._marg.keys(), \
        "Key {} not marginal in distribution {}".format(key, self._name)
    keys = set(keys)
    sum_axes = []
    dims = collections.OrderedDict(self._dims)
    for i, key in enumerate(self._keylist):
      if key in keys:
        if self._dims[key] is not None:
          sum_axes.append(self._dims[key])
        dims[key] = None
    prob = rescale(self._prob, self._pscale, 1.)
    if sum_axes:
      sum_prob = np.sum(prob, axis=tuple(set(sum_axes)), keepdims=False)
    else:
      sum_prob = np.sum(prob)
    vals = collections.OrderedDict()
    for i, key in enumerate(self._keylist):
      if key in keys:
        val = self[key] if not exponent else self[key]**exponent
        if self._aresingleton[i]:
          vals.update({key: val})
        else:
          expt_numerator = np.sum(prob*np.array(val, dtype=float), 
                                  axis=tuple(set(sum_axes)), keepdims=False)
          vals.update({key: div_prob(expt_numerator, sum_prob)})
      elif key in self._cond.keys():
        vals.update({key: self[key]})
    return vals

#-------------------------------------------------------------------------------
  def quantile(self, q=0.5):
    """ Returns probability quantiles in distribution for sorted values """
    quants = [q] if isscalar(q) else q

    # Deal with trivial scalar case
    if self._issingleton:
      quantiles = [collections.OrderedDict(self)] * len(quants)
      if isscalar(q):
        return quantiles[0]
      return quantiles

    # Check for monotonicity
    unsorted = set()
    for i, key in enumerate(self._keylist):
      if not self._aresingleton[i]:
        if not ismonotonic(self[key]):
          unsorted.add(key)

    # Evaluate quantiles from cumulative probability
    ravprob = rescale(np.ravel(self.prob), self._pscale, 1.)
    cumprob = np.cumsum(ravprob)
    cumprob = div_prob(cumprob, cumprob[-1])
    cum_idx = np.maximum(0, np.digitize(np.array(quants), cumprob)-1).tolist()

    # Interpolate in last axis
    quantiles = [None] * len(cum_idx)
    for j, _cum_idx in enumerate(cum_idx):
      rav_idx = int(_cum_idx)
      unr_idx = np.unravel_index(rav_idx, self._shape)
      quantiles[j] = collections.OrderedDict()
      for i, key in enumerate(self._keylist):
        if self._aresingleton[i]:
          quantiles[j].update({key: self[key]})
        elif key in unsorted:
          quantiles[j].update({key: {self[key].size}})
        else:
          dim = self._dims[key]
          val = np.ravel(self[key])
          idx = np.minimum(unr_idx[dim], len(val) - 1)
          if dim < self.ndim - 1 or idx == len(val) - 1:
            quantiles[j].update({key: val[idx]})
          else:
            vals = val[idx:idx+2]
            ravp = ravprob[rav_idx:rav_idx+2]
            if np.abs(np.diff(ravp)) < min(quants[j], 1. - quants[j]):
              cump = cumprob[rav_idx:rav_idx+2]
              interp_val = np.interp(quants[j], cump, vals)
              quantiles[j].update({key: interp_val})
            else:
              weighted_val = np.sum(ravp * vals) / np.sum(ravp)
              quantiles[j].update({key: weighted_val})
    if isscalar(q):
      return quantiles[0]
    return quantiles

#-------------------------------------------------------------------------------
  def sorted(self, key):
    """ Returns a distribution ordered by key """
    dim = self._dims[key]

    # Handle trivial case
    if dim is None:
      return PD(self._name, dict(self), dims=self._dims, 
                prob=self._prob, pscale=self._pscale)

    # Argsort needed to sort probabilities
    ravals = np.ravel(self[key])
    keyidx = np.argsort(ravals)
    keyval = ravals[keyidx]

    # Reorder probabilities in affected dimension
    slices = [slice(i) for i in self._shape]
    slices[dim] = keyidx
    prob = self._prob[tuple(slices)]

    # Evaluate values arrays
    vals = collections.OrderedDict()
    for _key, _val in self.items():
      if self._dims[_key] != dim:
        vals.update({_key: _val})
      elif _key == key:
        vals.update({_key: keyval})
      else:
        vals.update({_key: np.ravel(_val)[keyidx]})
    return PD(self.name, vals, dims=self._dims, 
              prob=prob, pscale=self._pscale)
    
#-------------------------------------------------------------------------------
  def rescaled(self, pscale=None):
    prob = rescale(np.copy(self.prob), self._pscale, pscale)
    return PD(self.name, dict(self), dims=self.dims,
              prob=prob, pscale=pscale)

#-------------------------------------------------------------------------------
  def __call__(self, values, keepdims=False):
    # Slices distribution according to scalar values given as a dictionary

    if isinstance(values, str):
      assert values in ['marg', 'cond'], \
          "String call must be either 'marg' or 'cond'"
      dictionary = self.marg if values == 'marg' else self.cond
      return self.lookup(list(dictionary.keys()))
    assert isinstance(values, dict),\
        "Values must be dict type, not {}".format(type(values))
    keys = values.keys()
    keyset = set(values.keys())
    assert len(keyset.union(self._keyset)) == len(self._keyset),\
        "Unrecognised key among values keys: {}".format(keys())
    marg = collections.OrderedDict(self._marg)
    cond = collections.OrderedDict(self._cond)
    dims = collections.OrderedDict(self._dims)
    vals = collections.OrderedDict(self)
    slices = [slice(None) for _ in range(self.ndim)]
    dim_delta = 0
    for i, key in enumerate(self._keylist):
      check_dims = False
      if not self._aresingleton[i]:
        dim = self.dims[key]
        if key in keyset:
          assert np.isscalar(values[key]), \
              "Values must contain scalars but found {} for {}".\
              format(values[key], key)
          match = np.ravel(self[key]) == values[key]
          n_matches = match.sum()
          post_eq = '{}'.format(values[key])
          if n_matches == 0:
            slices[dim] = slice(0, 0)
            vals[key] = np.array([])
          elif n_matches == 1 and not keepdims:
            dim_delta += 1
            slices[dim] = int(np.nonzero(match)[0])
            vals[key] = values[key]
            dims[key] = None
          else:
            post_eq = '[]'
            slices[dim] = np.nonzero(match)[0]
            vals[key] = self[key][slices[dim]]
            check_dims = True
          update_keys = [key]
          if check_dims:
            for k, v in self.items():
              if dim == self._dims[k] and k not in keyset:
                vals[k] = self[k][slices[dim]]
                update_keys.append(k)
          for update_key in update_keys:
            if key in marg.keys():
              marg[key] = "{}={}".format(key, post_eq)
            elif key in cond.keys():
              cond[key] = "{}={}".format(key, post_eq)
        elif dim_delta:
          dims[key] = dims[key] - dim_delta
    name = margcond_str(marg, cond)
    prob = self._prob[tuple(slices)]
    return PD(name, vals, dims=dims, prob=prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def __mul__(self, other):
    return product(*tuple([self, other]))

#-------------------------------------------------------------------------------
  def __add__(self, other):
    return summate(*tuple([self, other]))

#-------------------------------------------------------------------------------
  def __truediv__(self, other):
    """ If self is P(A, B | C, D), and other is P(A | C, D), this function
    returns P(B | C, D, A) subject to the following conditions:
    The divisor must be a scalar.
    The conditionals must match.
    The scalar marginals must match.
    """
    # Assert scalar division and operands compatible
    assert set(self.cond.keys())== set(other.cond.keys()),  \
      "Conditionals must match"

    divs = other.issingleton
    if divs:
      marg_scalars = set()
      for i, key in enumerate(self._keylist):
        if key in self._marg.keys() and self._aresingleton[i]:
          marg_scalars.add(key)
      assert marg_scalars == set(other.marg.keys()), \
        "For divisor singletons, scalar marginals must match"

    # Prepare quotient marg and cond keys
    keys = other.marg.keys()
    marg = collections.OrderedDict(self._marg)
    cond = collections.OrderedDict(self._cond)
    vals = collections.OrderedDict(self._cond)
    re_shape = np.ones(self._ndim, dtype=int)
    for i, key in enumerate(self._keylist):
      if key in keys:
        cond.update({key:marg.pop(key)})
        if not self._aresingleton[i] and not divs:
          re_shape[self.dims[key]] = other[key].size
      else:
        vals.update({key:self[key]})

    # Append the marginalised variables and end of vals
    for i, key in enumerate(self._keylist):
      if key in keys:
        vals.update({key:self[key]})

    # Evaluate probabilities
    name = margcond_str(marg, cond)
    divp = other.prob if divs else other.prob.reshape(re_shape)
    prob = div_prob(self._prob, divp, self._pscale, other.pscale)
    return PD(name, vals, dims=self._dims, prob=prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def remarginalise(self, distribution, *args, **kwds):
    """ Redistributes probability distribution within distribution according to a
    dictionary (in args[0]) or keywords, for which the keys represent the
    marginal keys and corresponding values contain their corresponding values
    according to the shape of self.prob. Conditional variables are unchanged.
    """

    # Simple assertions
    assert not self._issingleton,\
        "Function dist.remarginalise() not operative for scalar probabilities"
    assert isinstance(distribution, Distribution),\
        "First argument must be Distribution type not {}".format(type(distribution))

    # Read and check dictionary
    vals = None
    if args:
      assert len(args) == 1 and not kwds, \
          "Use a single dictionary argument or keywords, not both"
      vals = args[0]
    elif kwds:
      vals = dict(kwds)
    assert isinstance(vals, dict), \
        "Dictionary argument expected, not {}".format(type(vals))
    marg_keys = list(distribution.keys())
    dims = distribution.dims
    for key, val in vals.items():
      if dims[key] is not None:
        assert np.all(np.array(val.shape) == np.array(self.shape)),\
            "Values for {} incommensurate".format
        assert key in marg_keys, \
            "Key {} not present in inputted distribution"

    # Evaluate output indices for each space in probability
    shape = distribution.shape
    indices = [np.empty(_size, dtype=int) for _size in shape]
    for key, dim in dims.items():
      if dim is not None:
        edges = np.array(np.ravel(distribution[key]), dtype=float)
        rav_vals = np.array(np.ravel(vals[key]), dtype=float)
        indices[dim] = np.maximum(0, np.minimum(len(edges)-1, \
                                     np.digitize(rav_vals, edges)-1))

    # Iterate through every self.prob value
    rav_prob = np.ravel(rescale(self._prob, self._pscale, 1.))
    rem_prob = np.zeros(distribution.shape, dtype=float)
    for i in range(self.size):
      rem_idx = tuple([idx[i] for idx in indices])
      rem_prob[rem_idx] += rav_prob[i]

    # Replace marginal keys and keep conditional keys
    rem_vals = collections.OrderedDict(distribution)
    rem_dims = collections.OrderedDict(distribution.dims)
    marg_str = []
    for key, val in distribution.items():
      marg_str.append(key)
      if val is not None:
        if np.isscalar(val):
          marg_str[-1] = "{}={}".format(key, val)
        else:
          marg_str[-1] = key + "=[]"
    rem_name = ','.join(marg_str)
    if '|' in self._name:
      rem_name += self._name.split('|')[1]
    cond_keys = list(self._cond.keys())
    if cond_keys:
      for key in cond_keys:
        rem_vals.update({key: self[key]})
        rem_dims.update({key: self._dims[key]})

    return PD(rem_name, rem_vals, dims=rem_dims,
              prob=rescale(rem_prob, 1., self._pscale),
              pscale=self._pscale)

#-------------------------------------------------------------------------------
  def __repr__(self):
    prefix = 'logp' if iscomplex(self._pscale) else 'p'
    suffix = '' if not self._issingleton else '={}'.format(self.prob)
    return prefix + "(" + self._name + ")" + suffix

#-------------------------------------------------------------------------------
  def serialise(self):
    serialised = super().serialise()
    short_name = self.short_name
    serialised[short_name].update({'prob': self._prob})
    serialised[short_name].update({'pscale': self.pscale})
    return serialised

#-------------------------------------------------------------------------------



""" Expr representation wrapping SymPy's Expr. """

#-------------------------------------------------------------------------------
import collections
import sympy
import numpy as np
from sympy.utilities.autowrap import ufuncify
from probayes.variable_utils import parse_as_str_dict

DEFAULT_EFUNS = [
                  ({np.ndarray: ufuncify}, [], {'backend':'numpy'})
                  ,
                  ({np.ndarray: sympy.lambdify}, ['numpy', 'scipy'], {} )
                ]

#-------------------------------------------------------------------------------
def collate_symbols(expr):
  """ Collates symbols from expression expr as an ordered dictionary """
  assert isinstance(expr, sympy.Expr), \
    "Input type not Expr type but: {}".format(expr)
  symbols = collections.OrderedDict()
  for putative_symbol in expr.free_symbols:
    symbol = putative_symbol
    while hasattr(symbol, 'symbol') and \
        hasattr(symbol, 'name'):
      symbol = symbol.symbol
    if hasattr(symbol, 'name'):
      symbols.update({symbol.name: symbol})
  return symbols

#-------------------------------------------------------------------------------
def sort_ordered_dict(ordereddict, order, reverse=False):
  """ Resorts a an ordered directionary by order (Nones are ignored) """
  assert isinstance(ordereddict, collections.OrderedDict), \
      "First input must be an ordered dictionary"
  keys = list(ordereddict.keys())
  if isinstance(order, (tuple, list)):
    assert len(ordereddict) == len(order), \
        "Inputs not commensurate"
    order = {key: val for key, val in zip(keys, order)}
  else:
    assert set(keys) == set(order.keys()), \
        "Mismatch between keys and order: {} vs {}".format(keys, order.keys())
  not_none_keys = [key for key in keys if order[key] is not None]
  if len(not_none_keys) < 2:
    return ordereddict
  not_none_vals = [order[key] for key in not_none_keys]
  sort_order = np.argsort(not_none_vals)
  sort_order = sort_order.tolist() if not reverse else \
               sort_order[::-1].tolist()
  if np.min(np.diff(sort_order)) > 0:
    return ordereddict
  sorted_keys = [not_none_keys[index] for index in sort_order]
  index = -1 
  sorted_dict = collections.OrderedDict()
  for i, key in enumerate(keys):
    if key not in not_none_keys:
      sorted_dict.update({key: ordereddict[key]})
    else:
      index += 1
      sorted_key = sorted_keys[index]
      sorted_dict.update({sorted_key: ordereddict[sorted_key]})
  return sorted_dict

#-------------------------------------------------------------------------------
class Expr:
  """ This class wraps sy.Expr. Sympy's dependence on __new__ to return
  modified class objects at instantiation doesn't play nicely with multiple
  inheritance wrap them in here as a class instead and copy over the attributes.

  :example
  >>> import sympy as sy
  >>> from probayes.expr import Expr
  >>> x = sy.Symbol('x')
  >>> x_inc = Expr(x + 1) >>> print(x_inc.subs({x:1}))
  2
  >>>
  """

  # Public
  _expr = None    # Expression object

  # Protected
  _symbols = None # Ordered dictionary of symbols keyed by name
  _efun = None    # Dictionary of evaluation functions
  _remap = None   # Optional dictionary for remapping

#-------------------------------------------------------------------------------
  def __init__(self, expr, **kwds):
    self.expr = expr
    self.remap = None if 'remap' not in kwds else kwds.pop('remap')

#-------------------------------------------------------------------------------
  @property
  def expr(self):
    return self._expr

  @property
  def symbols(self):
    return self._symbols

  @expr.setter
  def expr(self, expr=None):
    """ Sets the expr object for this instance. Either pass a sy.Expr object 
    directly or in accordance with the calling conventions for sy.Expr.__new__
    """
    self._expr = expr
    self._symbols = collate_symbols(self._expr)

    # Make instance play nicely with Sympy by copying attributes and hash content
    members = dir(self._expr)
    for member in members:
      if not hasattr(self, member):
        try:
          attribute = getattr(self._expr, member)
          setattr(self, member, attribute)
        except AttributeError:
          pass

    # In addition to substitution, add default evaluation functions
    for efun in DEFAULT_EFUNS:
      self.add_efun(efun[0], *tuple(efun[1]), **dict(efun[2]))

#-------------------------------------------------------------------------------
  @property
  def remap(self):
    return self._remap

  @remap.setter
  def remap(self, remap=None):
    self._remap = remap

#-------------------------------------------------------------------------------
  @property
  def efun(self):
    return self._efun

  def add_efun(self, efun=None, *args, **kwds):
    """ Adds an expression evaluation function.
    :param efun: a single dictionary entry in the form {type: function}
    :param args: optional args to pass to efun
    :param kwds: optional kwds to pass to efun

    :returns: updated efun dictionary.
    """
    if self._efun is None:
      self._efun = {None: (self._expr.subs)}
    if efun is None or not self._symbols:
      return
    symbols = tuple(self._symbols.values())
    assert isinstance(efun, dict), \
        "Input efun must be dictionary type, not {}".format(type(efun))
    for key, val in efun.items():
      if key not in self._efun:
        try:
          self._efun.update({key: val(symbols, self._expr, *args, **kwds)})
        except (TypeError, ImportError, AttributeError): # Bypass sympy bugs
          pass
          # import pdb; pdb.set_trace()
    return self._efun

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
    """ Performs evaluation of expression inputting symbols as a dictionary 
    (see sympy.Symbol.subs() input specificion using dictionaries. """
    parse_values = True
    keys = list(self._symbols.keys())
    if not(len(kwds)) and len(args) == len(self._symbols):
      if not any([isinstance(arg, dict) for arg in args]):
        values = collections.OrderedDict()
        parse_values = False
        for i, key in enumerate(self._symbols.keys()):
          values.update({key: args[i]})
    if parse_values:
      if self._remap:
        kwds = dict(kwds)
        kwds.update({'remap': self.remap})
      values = parse_as_str_dict(*args, **kwds)

    # While determining type, collect evalues in required order
    etype = None
    evalues = [None] * len(keys)
    ekeys = [None] * len(keys)
    isarray = [None] * len(keys)
    vec_dim = [None] * len(keys)
    for i, key in enumerate(keys):
      if isinstance(values[key], bool): # Sympy doesn't like bool types
        values[key] = int(values[key])
      ekeys[i] = self._symbols[key]
      evalues[i] = values[key]
      _etype = type(evalues[i])
      isarray[i] = _etype is np.ndarray
      if isarray[i]:
        non_sing = np.array(evalues[i].shape) > 1
        non_sing_sum = non_sing.sum()
        if non_sing_sum > 0:
          vec_dim[i] = int(np.nonzero(non_sing)[0][0])
      if etype is None: 
        if _etype in self._efun.keys():
          etype = _etype
      elif _etype in self._efun.keys():
        assert etype == _etype, \
            "Cannot mix types {} vs {} ".format(etype, _etype)

    # If supported efun found or not array, then use it
    if not(any(isarray)) or np.ndarray in self._efun:
      efun = self._efun[etype]
      vals = efun(*evalues) if etype else efun(values)
      if isinstance(vals, np.ndarray):
        return vals
      elif isinstance(vals, sympy.Integer):
        return int(vals)
      elif isinstance(vals, sympy.Float):
        return float(vals)
      elif isinstance(vals, sympy.Expr) and not len(collate_symbols(vals)):
        return float(vals)
      elif etype == np.ndarray:
        return np.array(vals)
      return vals

    # Try to support iterating numpy-array even without a NumPy eval func
    if any(isarray) and np.ndarray not in self._efun:
      efun = self._efun[None]

      vec_dims = np.array([dim for dim in vec_dim if dim is not None])
      # Is a sort required (meshgrid seems to prefer reverse order)?
      if len(vec_dims) > 1 and np.max(np.diff(vec_dims)) > 0:
        values = sort_ordered_dict(values, vec_dim, reverse=True)
        evalues = list(values.values())
      mesh_values = [val for i, val in enumerate(evalues) if isarray[i]]
      mesh_grid = np.meshgrid(*tuple(mesh_values), copy=False)
      revalues = collections.OrderedDict()
      shape = None
      mesh_idx = -1
      for i, key in enumerate(keys):
        if isarray[i]:
          mesh_idx += 1
          shape = mesh_grid[mesh_idx].shape
          revalues.update({key: np.ravel(mesh_grid[mesh_idx])})
        else:
          revalues.update({key: values[key]})
      size = int(np.prod(shape))
      vals = dict(values)
      reval = [None] * size
      for k in range(size):
        for i, key in enumerate(keys):
          if isarray[i]:
            vals.update({key: revalues[key][k]})
        val = efun(vals)
        if isinstance(val, sympy.Integer):
          reval[k] = int(val)
        else:
          reval[k] = float(val)
      return np.array(reval).reshape(shape)

    # Otherwise rely on evaluation function

#-------------------------------------------------------------------------------
  def __repr__(self):
    if self._expr is None:
      return super().__repr__()
    return self._expr.__repr__()

#-------------------------------------------------------------------------------
  def __inv__(self):
    """ Overload bitwise inverter operator to return symbol. """
    return self.symbol

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
    """ Returns the expression object if arg is slice(None) i.e. [:] """
    assert arg == slice(None), \
        "Expr[arg] must bec called using slice operator i.e. Expr[:], not: " + \
        "{}".format(arg)
    return self._expr

#-------------------------------------------------------------------------------
  def subre(self, expr=None):
    """ Substitutes known symbols within expression with explicitly real
    equivalents, denoted using __name__. """
    expr = expr or self._expr
    real_symbols = collections.OrderedDict()
    for key, val in self._symbols.items():
      real_name = '__{}__'.format(key)
      real_symbols.update({val: sympy.Symbol(real_name, real=True)})
    return expr.subs(real_symbols)

#-------------------------------------------------------------------------------
  def resub(self, real_expr):
    """ Reverses the substitution performed by subre. """
    real_symbols = collate_symbols(real_expr)
    symbols = collections.OrderedDict()
    for key, val in real_symbols.items():
      if key[:2] == '__' and key[-2:] == '__':
        orig_name = key[2:-2]
        if orig_name in self._symbols:
          symbols.update({val: self._symbols[orig_name]})
    return real_expr.subs(symbols)

#-------------------------------------------------------------------------------
  def simplify(self, expr=None):
    """ Simplifies expr via conversion of known symbols to real """
    simplified = self.subre(expr)
    return self.resub(simplified)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()

#------------------------------------------------------------------------------



""" Provides operator functions for deterministic and stochastic classes """

#-------------------------------------------------------------------------------
def and_op(self, other):
  """ Unites (self, other) for Variables, Fields, Dependences, and their
  stochastic equivalents. """
  from probayes.variable import Variable
  from probayes.field import Field
  from probayes.rf import RF
  from probayes.dependence import Dependence
  from probayes.sd import SD

  # Determine whether stochastic or non-stochastic field
  assert hasattr(self, 'is_stochastic'), \
      "Unrecognised pre-operand type {}".format(type(self))
  assert hasattr(other, 'is_stochastic'), \
      "Unrecognised post-operand type {}".format(type(other))
  fld, dep = Field, Dependence
  if any([self.is_stochastic, other.is_stochastic]):
    fld, dep = RF, SD

  # Dependence as pre-operand
  if isinstance(self, Dependence):
    return dep(other, self)

  # Field as pre-operand
  if isinstance(self, Field):
    if isinstance(other, Dependence):
      leafs = list(self.leafs.varlist) + \
              list(other.leafs.varlist)
      stems = other.stems
      roots = other.roots
      args = [fld(*tuple(leafs))]
      if stems:
        args += list(stems.values())
      if roots:
        args += [roots]
      return dep(*tuple(args))

    if isinstance(other, Field):
      leafs = list(self.vars.values()) + list(other.vars.values())
      return fld(*tuple(leafs))

    if isinstance(other, Variable):
      leafs = list(self.vars.values()) + [other]
      return fld(*tuple(leafs))

    raise TypeError("Unrecognised post-operand type {}".format(type(other)))


  # Variable as pre-operand
  if isinstance(self, Variable):
    if isinstance(other, Dependence):
      leafs = [self] + list(other.leafs.vars.values())
      stems = other.stems
      roots = other.roots
      args = fld(*tuple(leafs))
      if stems:
        args += list(stems.values())
      if roots:
        args += list(roots.values())
      return dep(*args)

    if isinstance(other, Field):
      var = [self] + list(other.vars.values())
      return fld(*tuple(var))

    if isinstance(other, Variable):
      return fld(self, other)
  
    raise TypeError("Unrecognised post-operand type: ".format(type(other)))

  raise TypeError("Unrecognised pre-operand type: ".format(type(self)))

#-------------------------------------------------------------------------------
def or_op(self, other):
  """ Conditions self w.r.y. other for Variables, Fields, Dependences, and their
  stochastic equivalents. """
  from probayes.dependence import Dependence
  from probayes.sd import SD

  # Determine whether stochastic or non-stochastic dependence
  assert hasattr(self, 'is_stochastic'), \
      "Unrecognised pre-operand type {}".format(type(self))
  assert hasattr(other, 'is_stochastic'), \
      "Unrecognised post-operand type {}".format(type(other))
  dep = SD if any([self.is_stochastic, other.is_stochastic]) else Dependence

  # Check for unviable cross conditionaliiation
  if isinstance(self, Dependence) and isinstance(other, Dependence):
    if self.roots and other.roots:
      raise ValueError("Cannot cross-conditionalise dependencies")
  return dep(self, other)

#-------------------------------------------------------------------------------




# This module is to be deprecated along with Func.
'''A conditional function is a functional wrapper that describes the 
dependence of an about output RF with respect to an input RF or the
by the dependence of a subgroup of RVs with respect to the others. 
'''
import collections
from probayes.rv import RV
from probayes.func import Func

#-------------------------------------------------------------------------------
def _to_rf(obj=None):
  """ Converts obj (which may be an RV/RF) or list/tuple of such to an RF """
  from probayes.rf import RF
  if obj is None:
    return obj
  if isinstance(obj, RF):
    assert not obj.stems and not obj.roots, \
        "RFs must be rootless and stemless"
    return obj
  elif isinstance(obj, RV):
    return RF(obj)
  elif isinstance(obj, (list, tuple)):
    rvs = []
    for subobj in obj:
      if isinstance(subobj, RV):
        rvs += [subobj]
      elif isinstance(subobj, RF):
        assert not obj.stems and not obj.roots, \
            "RFs must be rootless and stemless"
        rvs += subobj.varlist
      else:
        raise TypeError("Unexpected type: {}".format(type(subobj)))
    return RF(*tuple(rvs))
  else:
    raise TypeError("Unexpected type: {}".format(type(subobj)))

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class CF (Func):
  '''A conditional function is a functional wrapper that describes the 
  dependence of an about output RF with respect to an input RF or the
  by the dependence of a subgroup of RVs with respect to the others. 
  '''
  # Protected
  _out = None
  _inp = None
  _name = None
  _passprob = None
  _passdims = None

  # Private
  __inpdict = None
  __unknown = None

#-------------------------------------------------------------------------------
  @property
  def callable(self):
    return self._Func__callable

  @property
  def isscipy(self):
    return self._Func__isscipy

  @property
  def isscalar(self):
    return self._Func__isscalar

  @property
  def ismulti(self):
    return self._Func__ismulti

  def __init__(self, out=None, inp=None, func=None, *args, **kwds):
    """ Sets the output and input random fields out and inp and initialises 
    the function according to object in func, which may be an uncallable
    object, a callable function, or a tuple of callable functions. See Func.
    """
    self.set_out(out)
    self.set_inp(inp)
    self.set_func(func, *args, **kwds)
    
#-------------------------------------------------------------------------------
  def set_out(self, out=None):
    """ Sets the output and input random fields out, which may be an RF an RV or 
    tuple or RFs and RVs but not an SD. 
    """
    self._out = _to_rf(out)

#-------------------------------------------------------------------------------
  def set_inp(self, inp=None):
    """ Sets the output and input random fields out and inp, each of which may 
    be an RF an RV or tuple or RFs and RVs but not an SD. Optionally inp may
    be an OrderedDict with RV names in the form (only keys are checked):
    {'x_0,x_1': 'x_2,x_3,x_4', 'x_2': 'x_0', 'x_3,x_4': 'x_0,x_1,x_2'}. 
    """
    self.__inpdict = isinstance(inp, (tuple, list, dict))
    if self.__inpdict:
      if isinstance(inp, (tuple, list)):
        inp = {key: '' for key in inp}
    self.__unknown = None

    # Handle inp is a random field first
    if not self.__inpdict and inp is not None:
      assert self._out is not None, "Set outputs before inputs"
      self._inp = _to_rf(inp)
      self._name = '|'.join([self._out.name, self._inp.name])
      return 
    elif inp is not None and not isinstance(inp, collections.OrderedDict):
      raise TypeError("Dictionary specification for inp must be an OrderedDict")

    # At this stage, there must be at least two RVs
    rv_keys = self._out.keylist
    assert len(rv_keys) > 1, "Multiple RVs required for conditional functions"

    # Check inp if entered
    if inp:
      all_inp_keys = [] 
      for key, val in inp.items():
        inp_keys = key.split(',')
        val_keys = val.split(',')
        for val_key in val_keys:
          if len(inp) > 1:
            assert val_key not in inp_keys, \
                "Circular key reference {}:{}".format(key, val)
          assert val_key in rv_keys, \
              "Key {} not recognised in RF {}".format(val_key, self._out)
        all_inp_keys += inp_keys
      assert set(rv_keys) == set(all_inp_keys), \
          "Conditional function {} specification incommensurate with RF {}".\
          format(inp, self._out)
      self._inp = inp
    prime_keys = ["{}'".format(rv_key) for rv_key in rv_keys]
    self._name = '|'.join([','.join(prime_keys), ','.join(rv_keys)])

#-------------------------------------------------------------------------------
  def set_func(self, func=None, *args, **kwds):
    """ Set the CF instance's function object.

    :param func: an uncallable object, callable function, or tuple of functions
    :param *args: arguments to pass onto callables
    :param **kwds: keywords to pass onto callables

    Note that the following two reserved keywords are disallowed:

    'order': which instead denotes a dictionary of remappings.
    'delta': which instead denotes a mapping of differences.
    'passdims': which provides an auxiliary flag to SD to pass dims
    'passprob': which provides an auxiliary flag to SD to pass prob`

    """

    self._func = func
    self._args = tuple(args)
    self._kwds = dict(kwds)
    self._passdims = None if 'passdims' not in kwds else self._kwds.pop('passdims')
    self._passprob = None if 'passprob' not in kwds else self._kwds.pop('passprob')
    if isinstance(self._func, Func) and not self._args and not self._kwds:
      func = self._func.ret_func()
      args = self._func.ret_args()
      kwds = self._func.ret_kwds()
      order = self._func.ret_order()
      delta = self._func.ret_delta()
      argout = super().set_func(func, *args, **kwds)
      self.set_order(order)
      self.set_delta(delta)
      return argout
    super().set_func(self._func, *self._args, **self._kwds)

#-------------------------------------------------------------------------------
  def ret_name(self):
    """ Return the conditional function name determined from the RFs """
    return self._name

#-------------------------------------------------------------------------------
  def ret_out(self):
    """ Returns output RF """
    return self._out

#-------------------------------------------------------------------------------
  def ret_inp(self):
    """ Returns input RF """
    return self._inp

#-------------------------------------------------------------------------------
  def ret_passdims(self):
    """ Returns passdims flag if specified """
    return self._passdims

#-------------------------------------------------------------------------------
  def ret_passprob(self):
    """ Returns passprob flag if specified """
    return self._passprob

#-------------------------------------------------------------------------------
  def ret_inpdict(self):
    """ Returns flag inpdict indicating whether a dict """
    return self.__inpdict

#-------------------------------------------------------------------------------
  def __getitem__(self, spec=None):
   r""" Calls the $i$th function from the Func tuple where the spec is $i$.
   If spec is a string, then the specification is changed.
   """
   if not isinstance(spec, str):
     return super().__getitem__(spec)
   assert self.__inpdict, \
       "Input spec as a string only supported with a OrderedDict inp"
   self.__unknown = spec
   return self._call

#-------------------------------------------------------------------------------
  def _call(self, *args, **kwds):
    if self.__unknown is None:
      return super()._call(*args, **kwds)
    kwds = dict(kwds)
    kwds.update({'unknown': self.__unknown})
    self.__unknown = None
    return super()._call(*args, **kwds)

#-------------------------------------------------------------------------------



# Utility module for SD objects

import collections
from probayes.pscales import prod_rule

#-------------------------------------------------------------------------------
def desuffix(values, suffix="'"):
  assert isinstance(values, dict), \
      "Input must be a dictionary, not {}".format(type(values))
  suffix_found = any([key[-1] == suffix for key in values.keys()])
  vals = collections.OrderedDict()
  if not suffix_found:
    vals.update({values})
    return vals
  for key, val in values.items():
    vals_key = key if key[-1] != suffix else key[:-1]
    assert vals_key not in vals, "Repeated key: {}".format(vals_key)
    vals.update({vals_key: val})
  return vals
  
#-------------------------------------------------------------------------------
def get_suffixed(values, unsuffix=True, suffix="'"):
  assert isinstance(values, dict), \
      "Input must be a dictionary, not {}".format(type(values))
  vals = collections.OrderedDict()
  for key, val in values.items():
    if key[-1] == suffix:
      vals_key = key[:-1] if unsuffix else key
      vals.update({vals_key: val})
  return vals

#-------------------------------------------------------------------------------
def arch_prob(arch, dims, **kwds):
  """ Returns the combined probability of for arch given values """
 
  values = dict(kwds)
  dimkeys = list(dims.keys())
  assert isinstance(arch, (tuple, list)), "Archictecture must be tuple or list"
  serial = isinstance(arch, list)
  probs = [None] * len(arch)
  for i, subarch in enumerate(arch):
    keyset = subarch.keylist
    vals = {key: values[key] for key in dimkeys if key in keyset}
    subdims = {key: dims[key] for key in dimkeys if key in keyset}
    probs[i] = subarch.eval_prob(vals, subdims)
  if serial:
    return probs[-1]
  pscales = [subarch.pscale for subarch in arch]
  prob, pscale = prod_rule(*tuple(probs), pscales=pscales)
  return prob

#-------------------------------------------------------------------------------



"""
A covariance-matrix-based conditional sampling class.
"""
import numpy as np
import scipy.stats
from probayes.vtypes import isunitsetint, uniform

#-------------------------------------------------------------------------------
class CondCov:

  # Protected
  _mean = None # Means 
  _cov = None  # Covariance matrix
  _n = None    # len(means)
  _scom = None # Schur complements
  _stdv = None # Conditional standard deviations
  _coef = None # Regression coefficients
  _lims = None  # Recentered limits (Nx2 array)
  _cdfs = None  # cdf-limits

#-------------------------------------------------------------------------------
  def __init__(self, mean, cov, lims):
    self._mean = np.atleast_1d(mean)
    self._cov = np.atleast_2d(cov)
    self._inv = np.linalg.inv(self._cov)
    self._lims = np.atleast_2d(lims) - np.expand_dims(self._mean, -1)
    self._n = len(self._mean)
    assert len(self._cov) == self._n, \
        "Means and covariance matrix incommensurate"
    self._stdv = np.empty(self._n, dtype=float)
    self._coef = [None] * self._n
    for i in range(self._n):
      ll  = np.delete(self._cov[:, i].reshape([self._n, 1]), (i), axis=0)
      ru  = np.delete(self._cov[i, :].reshape([1, self._n]), (i), axis=1)
      cov = np.delete(np.delete(self._cov, (i), axis=1), (i), axis=0)
      self._coef[i] = ru.dot(np.linalg.inv(cov))
      self._stdv[i] = np.sqrt(self._cov[i, i] - float(self._coef[i].dot(ll)))
    self._cdfs = np.array([scipy.stats.norm.cdf(lim, loc=0., scale=self._stdv[i]) \
                          for i, lim in enumerate(self._lims)])

#-------------------------------------------------------------------------------
  def interp(self, *args, cond_pdf=False):
    # args in order of mean - one must be a unitsetint
    idx = None
    dmu = np.empty((self._n, 1), dtype=float)
    vals = [arg for arg in args]
    for i, arg in enumerate(args):
      if isunitsetint(arg):
        if idx is None:
          idx = i
        else:
          raise ValueError("Only one argument can be interpolated at a time")
      else:
        dmu[i] = arg - self._mean[i]
    assert idx is not None, "No variable specified for interpolation"
    dmu = np.delete(dmu, (idx), axis=0)
    lims = self._cdfs[idx]
    number = list(args[idx])[0]
    cdf = uniform(lims[0], lims[1], number)
    mean = self._mean[idx] + float(self._coef[idx].dot(dmu))
    stdv = self._stdv[idx]
    vals = scipy.stats.norm.ppf(cdf, loc=mean, scale=stdv)
    if not cond_pdf:
      return vals
    return vals, scipy.stats.norm.pdf(vals, loc=mean, scale=stdv)

#-------------------------------------------------------------------------------



""" Provides Functional to contain multiple expressions """

import collections
import functools
from probayes.functional_utils import NX_UNDIRECTED_GRAPH, collate_vertices,\
                                      parse_identifiers
from probayes.expression import Expression

#-------------------------------------------------------------------------------
class Functional:
  '''A Functional is container for multiple functions that define a
  deterministic relationship between one graph of vertices and another.
  '''
  _out  = None     # Output graph
  _out_dict = None # Output vertices
  _out_set = None  # Set of output vertices _inp  = None     # Input graph
  _inp_dict = None # Input vertices
  _inp_set = None  # Set of input vertices
  _args = None     # Default args
  _kwds = None     # Default kwds
  _deps = None     # Dependences linking inputs to outputs
  _funcs = None    # Expression instances with same keys as deps
  _ndeps = None    # len(deps)
  _isiconic = None # Flag for being iconic
  _partials = None # Dict of partial functions

#-------------------------------------------------------------------------------
  def __init__(self, out=None, inp=None, *args, **kwds):
    """ Initialises output and input objects for the functional, where out and
    inp are undirected graphs made of vertices comprising object instances with
    members node.name comprising an identifier string."""
    self.out = out
    self.inp = inp or self._out
    self._ndeps = 0
    self._args = tuple(args) 
    self._kwds = dict(kwds)

#-------------------------------------------------------------------------------
  @property
  def out(self):
    """ Output graph """
    return self._out

  @property
  def out_dict(self):
    """ Output dictionary """
    return self._out_dict

  @property
  def out_set(self):
    """ Output set """
    return self._out_set
  
  @out.setter
  def out(self, out):
    self._out = out
    self._out_dict = collate_vertices(self._out)
    self._out_set = frozenset(self._out_dict.keys())

#-------------------------------------------------------------------------------
  @property
  def inp(self):
    """ Input graph """
    return self._inp
  
  @property
  def inp_dict(self):
    """ Input dictionary """
    return self._inp_dict

  @property
  def inp_set(self):
    """ Input set """
    return self._inp_set
  
  @inp.setter
  def inp(self, inp):
    self._inp = inp
    self._inp_dict = collate_vertices(self._inp)
    self._inp_set = frozenset(self._inp_dict.keys())

#-------------------------------------------------------------------------------
  @property
  def deps(self):
    """ Input/output dependences dictionary """
    return self._deps

  def __len__(self):
    return self._ndeps

  def __bool__(self):
    if self._ndeps:
      return True
    return False

  @property
  def ndeps(self):
    return self._ndeps

  @property
  def funcs(self):
    """ Expressions dictionary """
    return self._funcs
  
  @property
  def isiconic(self):
    """ Iconic flag """
    return self._iconic

  def add_func(self, spec, func, *args, **kwds):
    """ Adds an function dependency for a specificied inp/out relationship where
    func, *args, **kwds are the inputs to the corresponding Expression instance
    (see Expression) and spec specifies the dependent variable(s) (for iconic
    expressions) or (for non-iconic) both dependent and variables.
    """

    # Initialise deps and funcs if not set
    if self._deps is None:
      self._deps = collections.OrderedDict()
    if self._funcs is None:
      self._funcs = collections.OrderedDict()

    # Update deps
    spec_out = None
    spec_inp = None
    if isinstance(spec, dict):
      spec_out = parse_identifiers(tuple(spec.keys()))
      spec_inp = parse_identifiers(tuple(spec.values()))
    else:
      spec_out = parse_identifiers(spec)

    if self._deps:
      assert spec_out not in self._deps.keys(), \
          "Output dependence for {} already previously set".format(spec_out)
    self._deps.update({spec_out: spec_inp})
    self._ndeps = len(self._deps)

    # Update funcs and set isiconic flag if not previously set
    self._funcs.update({spec_out: Expression(func, *args, **kwds)})
    if self._isiconic is None:
      self._isiconic = self._funcs[spec_out].isiconic
    else:
       assert self._isiconic == self._funcs[spec_out].isiconic, \
           "Cannot mix iconic and non-iconic expressions within functional"

    # If iconic, check for single spec_out and update dexpr if available
    if self._isiconic:
      assert not len(args) and not len(kwds), \
          "No optional arguments or keywords supported for iconic functions"
      assert len(spec_out) == 1, \
          "Only single output functions supported for iconic expressions"
      key = list(spec_out)[0]
      assert key in self._out_set, \
          "Key {} not found in amount output set {}".format(key, self._out_set)
      var = self._out_dict[key]
      if hasattr(var, 'dexpr'):
        var.dexpr = func

    # Detect inputs for iconics if not specified
    if spec_inp is None:
      assert self._isiconic, \
          "Input specification mandatory for non-iconic functionals"
      spec_inp = tuple(self._funcs[spec_out].symbols.keys())
      spec_inp = parse_identifiers(spec_inp)
      self._deps[spec_out] = spec_inp

    # Detect subsets are covered
    assert spec_out.issubset(self._out_set), \
        "Specification output must be a subset of output graph"
    assert spec_inp.issubset(self._inp_set), \
        "Specification input must be a subset of input graph"

    self._set_partials()

#-------------------------------------------------------------------------------
  @property
  def partials(self):
    return self._partials

  def _set_partials(self):
    # Protected function to update partial function dictionary of calls
    self._partials = collections.OrderedDict()
    for key, val in self._funcs.items():
      call = functools.partial(val.__call__, *self._args, **self._kwds)
      self._partials.update({key: call})
      if not val.order:
        order = dict()
        deps = self._deps[key]
        for element in self._inp_set:
          if element not in deps:
            order.update({element: None})
        val.set_order(order)

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
   """ Returns the partial according to arg  """ 
   key = parse_identifiers(arg)
   return self._partials[key]

#-------------------------------------------------------------------------------
  def __repr__(self):
    """ Print representation """
    if self._isiconic:
      return object.__repr__(self)+ ": '{}'".format(self._funcs)
    if self._deps is None:
      return object.__repr__(self) 
    return object.__repr__(self)+ ": '{}'".format(self._deps)

#-------------------------------------------------------------------------------



'''
A wrapper for sympy-based probability objectsj
Note we eschew sympy.stats.density etc...
'''
import collections
import functools
import numpy as np
import sympy
import sympy.stats
import scipy.stats
from probayes.expr import Expr


#-------------------------------------------------------------------------------
SYMPY_STATS_DIST = {sympy.stats.rv.RandomSymbol}
def is_sympy_stats_dist(arg, sympy_stats_dist=SYMPY_STATS_DIST):
  """ Returns if arguments belongs to sympy.stats.continuous or discrete """
  return isinstance(arg, tuple(sympy_stats_dist))

#-------------------------------------------------------------------------------
def sympy_obj_from_dist(dist, _recursive=False):
  """ Attempts to return the object class instance for sympy distribution.
  :example:
  >>> import sympy
  >>> import sympy.stats
  >>> import probayes as pb
  >>> x = sympy.Symbol('x')
  >>> p_x = sympy.stats.Normal(x, mean=0, std=1.)
  >>> p_x_obj = pb.sympy_obj_from_dist(p_x)
  >>> print(p_x_obj.pdf)
  0.5*sqrt(2)*exp(-0.5*x**2)/sqrt(pi)
  """
  obj = dist
  if hasattr(obj, '_cdf'):
    return obj
  if not hasattr(obj, 'args'):
    if not _recursive:
      raise ValueError('Unsupported distribution: {}'.format(dist))
    return None
  objs = [sympy_obj_from_dist(arg, True) for arg in obj.args]
  if any(objs):
    for obj in objs:
      if obj:
        return obj
  if not _recursive:
    raise ValueError('Unsupported distribution: {}'.format(dist))
  return None

#-------------------------------------------------------------------------------
def sympy_sfun(distr, size=0, dtype=None, _sfunc=sympy.stats.sample):
  """ Sampling function for Sympy distributions where:

  :param distr: Sympy stats distribution
  :param size: size specification (as per scipy rvs)
  :param dtype: Numpy-style dtype specification

  :return randomly samples from distribution.
  """
  if not size:
    samples = _sfunc(distr)
    if dtype:
      return (dtype)(samples)
    return samples
  if dtype:
    samples = [dtype(_sfunc(distr)) for _ in range(size)]
    return np.array(samples)
  samples = [_sfunc(distr) for _ in range(size)]
  return samples

#-------------------------------------------------------------------------------
def bernoulli_prob(var, bias=0.5):
  """ Returns Sympy-expression of a Bernoulli PMF for var at specified bias.

  :param var: boolean symbol to express pmf
  :param bias: bias probability (float ranging from 0 to 1).

  :return bias*var + (1-bias)*(1-var)
  """
  # bv + (1-b)(1-v) = bv + 1 - v - b + bv = 2bv - v - b + 1 = v(2b-1) + (1-b)
  return var * (2*bias - 1) + (1 - bias)

#-------------------------------------------------------------------------------
def bernoulli_sfun(size=None, bias=0.5):
  """ Returns Bernoulli random variates of given size for given bias """
  bern = scipy.stats.bernoulli(p=float(bias))
  if not size:
    return bool(bern.rvs())
  return bern.rvs(size).astype(bool)

#-------------------------------------------------------------------------------
class SympyProb:
  """ An expression wrapper for Sympy-based probabilities:

  :example

  >>> import sympy
  >>> import sympy.stats
  >>> import probayes as pb
  >>> import numpy as np
  >>> x = sympy.Symbol('x')
  >>> normal = pb.SympyProb(sympy.stats.Normal(x, mean=0, std=1))
  >>> print(normal['prob'](np.array([-1., 0., 1.])))
  [0.24197072 0.39894228 0.24197072]
  """

  # Protected
  _distr = None # Distribution function
  _probj = None # Probability distribution object
  _exprs = None # Distribution expression dictionary
  _cterm = None # Cumulative term for CDFs 
  __icdf = None # Inverse CDF symbol

#-------------------------------------------------------------------------------
  def __init__(self, distr=None):
    self.distr = distr
    
#-------------------------------------------------------------------------------
  @property
  def distr(self):
    return self._distr

  @property
  def probj(self):
    return self._probj

  @property
  def exprs(self):
    return self._exprs

  @distr.setter
  def distr(self, distr=None):
    """ Sets the distribution object. This must be a Sympy.stats distribution.
    """
    self._distr = distr
    self._probj = None
    self._exprs = None
    if self._distr is None:
      return
    assert is_sympy_stats_dist(self._distr), \
        "Input dist must a Sympy statistical distribution: {} entered".format(
            self._distr)
    self._probj = sympy_obj_from_dist(self._distr)
    self._cterm = self._distr.args[0] # assume first term is cumulative
    self._exprs = collections.OrderedDict()

    if not isinstance(self._cterm, sympy.Symbol):
      raise TypeError("Unexpected first argument type: {}".format(type(self._cterm)))
    
    # Set PDF, CDF, ICDF
    if hasattr(self._probj, 'pdf'):
      self._exprs.update({'prob': Expr(self._probj.pdf(self._cterm))})
      prob_expr = self._exprs['prob']
      log_prob = sympy.log(prob_expr.subre(prob_expr.expr)).simplify()
      log_prob= prob_expr.resub(log_prob)
      self._exprs.update({'logp': Expr(log_prob)})
    if hasattr(self._probj, '_cdf'):
      try:
        cdf_expr = Expr(self._probj._cdf(self._cterm))
        self._exprs.update({'cdf': cdf_expr})
      except:
        pass
    if 'cdf' in self._exprs.keys():
      icdf_name = "_{}_cdf".format(self._cterm.name)
      self.__icdf = sympy.Symbol(icdf_name)
      try:
        invexprs = sympy.solve(self._exprs['cdf'].expr - self.__icdf,  self._cterm)
      except TypeError: # in case Sympy fails to solve the inverse
        invexprs = []
      n_exprs = len(invexprs)
      if n_exprs:
        invexpr = invexprs[0]
        if n_exprs > 1:
          for expr in invexprs[1:]:
            if len(expr.__repr__()) < len(invexpr.__repr__()):
              invexpr = expr
        self._exprs.update({'icdf': Expr(invexpr, remap={self._cterm.name: icdf_name})})

    # Set sampling function
    self._exprs.update({'sfun': functools.partial(sympy_sfun, self._distr)})

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
    """ Returns class member or partials dictionary object according to arg:

    If None or [:], the entire expression dictionary is returned.
    If [], then the original distribution is returned.
    If {}, then the distribution probability class object is returned.
    If '', then the keys of the partials dictionary are returned.
    Otherwise arg is treated as the key for the expressions dictionary.
    """
    if isinstance(arg, tuple):
      raise NotImplementedError("Tuple input for __getitem__ not supported")
    if arg is None or arg == slice(None):
      return self._exprs
    if isinstance(arg, str) and not len(arg):
      return list(self._exprs.keys())
    if isinstance(arg, (list, dict)):
      assert not len(arg), "Input argument of type {} must be empty".type(arg)
      if isinstance(arg, list):
        return self._distr
      if isinstance(arg, dict):
        return self._probj
    return self._exprs[arg]

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()

#-------------------------------------------------------------------------------



"""
A Constants module. 
"""
import numpy as np

#-------------------------------------------------------------------------------
DEFAULT_FP_PRECISION = 64
COMPLEX_ZERO = complex(0., 0.)
FP_CONSTANTS = {32: {'nearly_positive_zero': 1.175494351e-38,
                     'nearly_positive_inf': 3.4022823466e38},
                64: {'nearly_positive_zero': 2.2250738585072014e-308,
                     'nearly_positive_inf': 1.7976931348623158e+308},
               }

NEARLY_POSITIVE_ZERO = None
NEARLY_POSITIVE_INF = None
NEARLY_NEGATIVE_INF = None
LOG_NEARLY_POSITIVE_INF =  None

def SET_FP_CONSTANTS(precision=DEFAULT_FP_PRECISION):
  """ Sets global floating point constants whether using precision=32 or
  precision=64 (default)
  """
  global NEARLY_POSITIVE_ZERO
  global NEARLY_POSITIVE_INF
  global NEARLY_NEGATIVE_INF
  global NEARLY_POSITIVE_INF
  global LOG_NEARLY_POSITIVE_INF
  NEARLY_POSITIVE_ZERO = FP_CONSTANTS[precision]['nearly_positive_zero']
  NEARLY_POSITIVE_INF = FP_CONSTANTS[precision]['nearly_positive_inf']
  NEARLY_NEGATIVE_INF = -NEARLY_POSITIVE_INF
  LOG_NEARLY_POSITIVE_INF = np.log(NEARLY_POSITIVE_INF)

if NEARLY_NEGATIVE_INF is None or LOG_NEARLY_POSITIVE_INF is None:
  SET_FP_CONSTANTS()

#-------------------------------------------------------------------------------




# A module to perform utility functions for PD() instances.

#-------------------------------------------------------------------------------
import collections
import functools
import numpy as np
import h5py
from probayes.vtypes import isscalar, issingleton, isunitsetint
from probayes.pscales import prod_pscale, prod_rule, iscomplex

#-------------------------------------------------------------------------------
def str2key(string):
  if isinstance(string, str):
    k = string.find('=')
    if k > 0:
      return string[:k]
    return string
  return [str2key(element) for element in string]

#-------------------------------------------------------------------------------
def str_margcond(name=None):
  """ Returns (marg,cond) tuple of OrderedDicts from a name """
  marg_str = name
  marg = collections.OrderedDict()
  cond = collections.OrderedDict()
  if not marg_str:
    return marg, cond
  lt_paren = marg_str.find('(')
  rt_paren = marg_str.find(')')
  if lt_paren >= 0 or rt_paren >= 0:
    assert lt_paren >= 0 and rt_paren > lt_paren, \
      "Unmatched parenthesis in name"
    marg_str = marg_str[lt_paren:1:rt_paren]
  cond_str = ''
  if '|' in marg_str:
    split_str = name.split('|')
    assert len(split_str) == 2, "Ambiguous name: {}".format(name)
    marg_str, cond_str = split_str
  marg_strs = []
  cond_strs = []
  if len(marg_str):
    marg_strs = marg_str.split(',') if ',' in marg_str else [marg_str] 
  if len(cond_str):
    cond_strs = cond_str.split(',') if ',' in cond_str else [cond_str]
  for string in marg_strs:
    marg.update({str2key(string): string})
  for string in cond_strs:
    cond.update({str2key(string): string})
  return marg, cond

#-------------------------------------------------------------------------------
def margcond_str(marg, cond):
  """ Returns a name from OrderedDict values in marg and cond """
  marg = list(marg.values()) if isinstance(marg, dict) else list(marg)
  cond = list(cond.values()) if isinstance(cond, dict) else list(cond)
  marg_str = ','.join(marg)
  cond_str = ','.join(cond)
  return '|'.join([marg_str, cond_str]) if cond_str else marg_str

#-------------------------------------------------------------------------------
def rekey_dict(ordereddict, keymap):
  assert isinstance(ordereddict, collections.OrderedDict), \
      "Marg or Cond dict objects must be Ordered Collections"
  assert isinstance(keymap, dict), \
      "Key map must be a dictionary"
  rekeyed = collections.OrderedDict()
  for key, val in ordereddict.items():
    if key not in keymap.keys():
      rekeyed.update({key: val})
    else:
      map_key = keymap[key]
      assert map_key not in rekeyed.keys(), \
          "Key remapping results in duplicate key: {}".format(map_key)
      string = val
      k = string.find('=')
      if k > 0:
        string = map_key + string[k:]
      else:
        string = map_key
      rekeyed.update({map_key: string})
  return rekeyed

#-------------------------------------------------------------------------------
def product(*args, **kwds):
  """ Multiplies two or more PDs subject to the following:
  1. They must not share the same marginal variables. 
  2. Conditional variables must be identical unless contained as marginal from
     another distribution.
  """
  from probayes.pd import PD

  # Check pscales, scalars, possible fasttrack
  if not len(args):
    return None
  kwds = dict(kwds)
  pscales = [arg.pscale for arg in args]
  pscale = kwds.get('pscale', None) or prod_pscale(pscales)
  aresingleton = [arg.issingleton for arg in args]
  maybe_fasttrack = all(aresingleton) and \
                    np.all(pscale == np.array(pscales)) and \
                    pscale in [0, 1.]


  # Collate vals, probs, marg_names, and cond_names as lists
  vals = [collections.OrderedDict(arg) for arg in args]
  probs = [arg.prob for arg in args]
  marg_names = [list(arg.marg.values()) for arg in args]
  cond_names = [list(arg.cond.values()) for arg in args]

  # Detect uniqueness in marginal keys and identical conditionals
  all_marg_keys = []
  for arg in args:
    all_marg_keys.extend(list(arg.marg.keys()))
  marg_sets = None
  if len(all_marg_keys) != len(set(all_marg_keys)):
    marg_keys, cond_keys, marg_sets, = None, None, None
    for arg in args:
      if marg_keys is None:
        marg_keys = list(arg.marg.keys())
      elif marg_keys != list(arg.marg.keys()):
        marg_keys = None
        break
      if cond_keys is None:
        cond_keys = list(arg.cond.keys())
      elif cond_keys != list(arg.cond.keys()):
        marg_keys = None
        break
      if marg_keys:  
        are_marg_sets = np.array([isunitsetint(arg[marg_key]) for
                                  marg_key in marg_keys])
        if marg_sets is None:
          if np.any(are_marg_sets):
            marg_sets = are_marg_sets
          else:
            marg_keys = None
            break
        elif not np.all(marg_sets == are_marg_sets):
          marg_keys = None
          break
    assert marg_keys is not None and marg_sets is not None, \
      "Non-unique marginal variables for currently not supported: {}".\
      format(all_marg_keys)
    maybe_fasttrack = True

  # Maybe fast-track identical conditionals
  if maybe_fasttrack:
    marg_same = True
    cond_same = True
    if marg_sets is None: # no need to recheck if not None (I think)
      marg_same = True
      for name in marg_names[1:]:
        if marg_names[0] != name:
          marg_same = False
          break
      cond_same = not any(cond_names)
      if not cond_same:
        cond_same = True
        for name in cond_names[1:]:
          if cond_names[0] != name:
            cond_same = False
            break
    if marg_same and cond_same:
      marg_names = marg_names[0]
      cond_names = cond_names[0]
      prod_marg_name = ','.join(marg_names)
      prod_cond_name = ','.join(cond_names)
      prod_name = '|'.join([prod_marg_name, prod_cond_name])
      prod_vals = collections.OrderedDict()
      for i, val in enumerate(vals):
        areunitsetints = np.array([isunitsetint(_val) 
                                   for _val in val.values()])
        if not np.any(areunitsetints):
          prod_vals.update(val)
        else:
          assert marg_sets is not None, "Variable mismatch"
          assert np.all(marg_sets == areunitsetints[:len(marg_sets)]), \
              "Variable mismatch"
          if not len(prod_vals):
            prod_vals.update(collections.OrderedDict(val))
          else:
            for j, key in enumerate(prod_vals.keys()):
              if areunitsetints[j]:
                prod_vals.update({key: {list(prod_vals[key])[0] + \
                                        list(val[key])[0]}})
      if marg_sets is not None:
        prob, pscale = prod_rule(*tuple(probs), pscales=pscales, pscale=pscale)
        return PD(prod_name, prod_vals, dims=args[0].dims, prob=prob, pscale=pscale)
      else:
        prod_prob = float(sum(probs)) if iscomplex(pscale) else float(np.prod(probs))
        return PD(prod_name, prod_vals, prob=prod_prob, pscale=pscale)

  # Check cond->marg accounts for all differences between conditionals
  prod_marg = [name for dist_marg_names in marg_names \
                          for name in dist_marg_names]
  prod_marg_name = ','.join(prod_marg)
  flat_cond_names = [name for dist_cond_names in cond_names \
                          for name in dist_cond_names]
  cond2marg = [cond_name for cond_name in flat_cond_names \
                         if cond_name in prod_marg]
  prod_cond = [cond_name for cond_name in flat_cond_names \
                         if cond_name not in cond2marg]
  cond2marg_set = set(cond2marg)

  # Check conditionals compatible
  prod_cond_set = set(prod_cond)
  cond2marg_dict = {name: None for name in cond2marg}
  for i, arg in enumerate(args):
    cond_set = set(cond_names[i]) - cond2marg_set
    if cond_set:
      assert prod_cond_set == cond_set, \
          "Incompatible product conditional {} for conditional set {}: ".format(
              prod_cond_set, cond_set)
    for name in cond2marg:
      if name in arg.keys():
        values = arg[name]
        if not isscalar(values):
          values = np.ravel(values)
        if cond2marg_dict[name] is None:
          cond2marg_dict[name] = values
        elif not np.allclose(cond2marg_dict[name], values):
          raise ValueError("Mismatch in values for condition {}".format(name))

  # Establish product name, values, and dimensions
  prod_keys = str2key(prod_marg + prod_cond)
  prod_nkeys = len(prod_keys)
  prod_aresingleton = np.zeros(prod_nkeys, dtype=bool)
  prod_areunitsetints = np.zeros(prod_nkeys, dtype=bool)
  prod_cond_name = ','.join(prod_cond)
  prod_name = prod_marg_name if not len(prod_cond_name) \
              else '|'.join([prod_marg_name, prod_cond_name])
  prod_vals = collections.OrderedDict()
  for i, key in enumerate(prod_keys):
    values = None
    for val in vals:
      if key in val.keys():
        values = val[key]
        prod_areunitsetints[i] = isunitsetint(val[key])
        if prod_areunitsetints[i]:
          values = {0}
        break
    assert values is not None, "Values for key {} not found".format(key)
    prod_aresingleton[i] = issingleton(values)
    prod_vals.update({key: values})
  if np.any(prod_areunitsetints):
    for i, key in enumerate(prod_keys):
      if prod_areunitsetints[i]:
        for val in vals:
          if key in val:
            assert isunitsetint(val[key]), "Mismatch in variables {} vs {}".\
                format(prod_vals, val)
            prod_vals.update({key: {list(prod_vals[key])[0] + list(val[key])[0]}})
  prod_newdims = np.array(np.logical_not(prod_aresingleton))
  dims_shared = False
  for arg in args:
    argdims = [dim for dim in arg.dims.values() if dim is not None]
    if len(argdims) != len(set(argdims)):
      dims_shared = True

  # Shared dimensions limit product dimensionality
  if dims_shared:
    seen_keys = set()
    for i, key in enumerate(prod_keys):
      if prod_newdims[i] and key not in seen_keys:
        for arg in args:
          if key in arg.dims:
            dim = arg.dims[key]
            seen_keys.add(key)
            for argkey, argdim in arg.dims.items():
              seen_keys.add(argkey)
              if argkey != key and argdim is not None:
                if dim == argdim:
                  index = prod_keys.index(argkey)
                  prod_newdims[index] = False

  prod_cdims = np.cumsum(prod_newdims)
  prod_ndims = prod_cdims[-1]

  # Fast-track scalar products
  if maybe_fasttrack and prod_ndims == 0:
     prob = float(sum(probs)) if iscomplex(pscale) else float(np.prod(probs))
     return PD(prod_name, prod_vals, prob=prob, pscale=pscale)

  # Reshape values - they require no axes swapping
  ones_ndims = np.ones(prod_ndims, dtype=int)
  prod_shape = np.ones(prod_ndims, dtype=int)
  scalarset = set()
  prod_dims = collections.OrderedDict()
  for i, key in enumerate(prod_keys):
    if prod_aresingleton[i]:
      scalarset.add(key)
    else:
      values = prod_vals[key]
      re_shape = np.copy(ones_ndims)
      dim = prod_cdims[i]-1
      prod_dims.update({key: dim})
      re_shape[dim] = values.size
      prod_shape[dim] = values.size
      prod_vals.update({key: values.reshape(re_shape)})
  
  # Match probability axes and shapes with axes swapping then reshaping
  for i in range(len(args)):
    prob = probs[i]
    if not isscalar(prob):
      dims = collections.OrderedDict()
      for key, val in args[i].dims.items():
        if val is not None:
          dims.update({val: prod_dims[key]})
      old_dims = []
      new_dims = []
      for key, val in dims.items():
        if key not in old_dims:
          old_dims.append(key)
          new_dims.append(val)
      if len(old_dims) > 1 and not old_dims == new_dims:
        max_dims_inc = max(new_dims) + 1
        while prob.ndim < max_dims_inc:
          prob = np.expand_dims(prob, -1)
        prob = np.moveaxis(prob, old_dims, new_dims)
      re_shape = np.copy(ones_ndims)
      for dim in new_dims:
        re_shape[dim] = prod_shape[dim]
      probs[i] = prob.reshape(re_shape)

  # Multiply the probabilities and output the result as a distribution instance
  prob, pscale = prod_rule(*tuple(probs), pscales=pscales, pscale=pscale)

  return PD(prod_name, prod_vals, dims=prod_dims, prob=prob, pscale=pscale)


#-------------------------------------------------------------------------------
def summate(*args):
  """ Quick and dirty concatenation """
  from probayes.pd import PD
  if not len(args):
    return None
  pscales = [arg.pscale for arg in args]
  vals = [dict(arg) for arg in args]
  probs = [arg.prob for arg in args]

  # Check pscales are the same
  pscale = pscales[0]
  for _pscale in pscales[1:]:
    assert pscale == _pscale, \
        "Cannot summate distributions with different pscales"

  # Check marginal and conditional keys
  marg_keys = list(args[0].marg.keys())
  cond_keys = list(args[0].cond.keys())
  for arg in args[1:]:
    assert marg_keys == list(arg.marg.keys()), \
      "Marginal variable names not identical across distributions: {}"
    assert cond_keys == list(arg.cond.keys()), \
      "Conditional variable names not identical across distributions: {}"
  sum_keys = marg_keys + cond_keys
  sum_name = ','.join(marg_keys)
  if cond_keys:
    sum_name += '|' + ','.join(cond_keys)

  # If all singleton, concatenate in dimension 0
  if all([arg.issingleton for arg in args]):
    unitsets = {key: isunitsetint(args[0][key]) for key in sum_keys}
    sum_dims = {key: None if unitsets[key] else 0 for key in sum_keys}
    sum_vals = {key: 0 if unitsets[key] else [] for key in sum_keys}
    sum_prob = []
    for arg in args:
      for key, val in arg.items():
        if unitsets[key]:
          assert isunitsetint(val), \
              "Cannot mix unspecified set and specified values"
          sum_vals[key] += list(val)[0]
        else:
          assert not isunitsetint(val), \
              "Cannot mix unspecified set and specified values"
          sum_vals[key].append(val)
      sum_prob.append(arg.prob)
    for key in sum_keys:
      if unitsets[key]:
        sum_vals[key] = {sum_vals[key]}
      else:
        sum_vals[key] = np.ravel(sum_vals[key])
    sum_prob = np.ravel(sum_prob)
    return PD(sum_name, sum_vals, dims=sum_dims, prob=sum_prob, pscale=pscale)

  # 2. all identical but in one dimension: concatenate in that dimension
  # TODO: fix the remaining code of this function below
  sum_vals = collections.OrderedDict(args[0])
  sum_dims = [None] * (len(args) - 1)
  for i, arg in enumerate(args):
    if i == 0:
      continue
    for key in marg_keys:
      if sum_dims[i-1] is not None:
        continue
      elif not arg.singleton(key):
        key_vals = arg[key]
        if key_vals.size == sum_vals[key].size:
          if np.allclose(key_vals, sum_vals[key]):
            continue
        sum_dims[i-1] = arg.dims[key]
  assert len(set(sum_dims)) > 1, "Cannot find unique concatenation axis"
  sum_dim = sum_dims[0]
  sum_dims = args[0].dims
  key = marg_keys[sum_dim]
  sum_prob = np.copy(probs[0])
  for i, val in enumerate(vals):
    if i == 0:
      continue
    sum_vals[key] = np.concatenate([sum_vals[key], val[key]], axis=sum_dim)
    sum_prob = np.concatenate([sum_prob, probs[i]], axis=sum_dim)
  return PD(sum_name, sum_vals, dims=sum_dims, prob=sum_prob, pscale=pscale)

#-------------------------------------------------------------------------------
def ismonotonic(vals):
  if issingleton(vals):
    return True
  vals = np.ravel(vals)
  is_ge = vals[1:] >= vals[:-1]
  if len(np.unique(is_ge)) == 1:
    return True
  return False

#-------------------------------------------------------------------------------
def iterdict(dicts):
  """ Converts a tuple/list of dicts to a dict of lists """
  return collections.OrderedDict(
      {key: [_dict[key] for _dict in dicts if key in _dict]
             for key in functools.reduce(set.union, [set(_dict.keys()) 
                                                     for _dict in dicts])}
                                )

#-------------------------------------------------------------------------------
def serialise(*args):
  from probayes import Distribution
  from probayes import PD
  serialised = {}
  for arg in args:
    if not isinstance(arg, (PD, Distribution)):
      raise TypeError(f"Unrecogised type to serialise: {type(arg)}")
    serialised.update(arg.serialise())
  return serialised

#-------------------------------------------------------------------------------
def deserialise(serialised):
  from probayes import Distribution
  from probayes import PD
  assert isinstance(serialised, dict), \
    f"Dict-type serialised input expected, not {type(serialised)}"
  dists = []
  for dist_name, _dist_dict in serialised.items():
    dist_dict = dict(_dist_dict)
    dims = {} if 'attrs' not in dist_dict else dist_dict.pop('attrs')
    prob = None if 'prob' not in dist_dict else dist_dict.pop('prob')
    pscale = None if 'pscale' not in dist_dict else dist_dict.pop('pscale')
    if prob is None:
      dists.append(Distribution(dist_name, dist_dict, dims=dims))
    else:
      pscale = pscale if pscale is None else np.atleast_1d(pscale).tolist()[0]
      dists.append(PD(dist_name, dist_dict, prob=prob, pscale=pscale, dims=dims))
  return tuple(dists)

#-------------------------------------------------------------------------------
def write_serialised(path, serialised, aux_dict={}):
  assert isinstance(serialised, dict), \
    f"Dict-type serialised input expected, not {type(serialised)}"
  for key, val in aux_dict.items():
    assert isinstance(key, str) and ' ' in key, \
      f"Aux dict must keyed by space-containing string, found {key}"
    assert isinstance(val, dict), \
      f"Aux dict must be a nested dictionary of dicts - found {type(val)}"
    for subkey, subval in val.items():
      assert isinstance(subkey, str), \
        f"Aux dict subkeys must be str, found {subkey}"
      assert isinstance(subvals, np.ndarray), \
        f"Aux dict subvals must be NumPy arrays, found {type(subvals)}"
  with h5py.File(path, 'w', libver='latest') as hdf_write:
    for dist_name, dist_dict in serialised.items():
      group_write = hdf_write.create_group(dist_name)
      for key, val in dist_dict.items():
        if key == 'attrs':
          attrs = dist_dict['attrs']
          for k, v in attrs.items():
            if v is None:
              attrs[k] = 'None'
          group_write[key] = np.array(len(attrs))
          group_write[key].attrs.update(attrs)
          if isinstance(attrs, collections.OrderedDict):
            group_write[key].attrs.update({'order': list(attrs.keys())})
        else:
          if isinstance(val, np.ndarray):
            group_write[key] = val
          elif isinstance(val, set):
            val_set = np.zeros(sorted(val) + [0], dtype='S')
            group_write[key] = val_set
          else:
            group_write[key] = np.array(val)
    for aux_name, aux_data in aux_dict.items():
      group_write = hdf_write.create_group(aux_name)
      for key, val in aux_data.items():
        group_write[key] = val

#-------------------------------------------------------------------------------
def read_serialised(path):
  serialised = {}
  aux_data = {}
  with h5py.File(path, 'r', libver='latest') as hdf_read:
    dist_names = hdf_read.keys()
    for dist_name in dist_names:
      group_read = hdf_read[dist_name]
      order = None
      attrs = {}
      if ' ' not in dist_name:
        serialised[dist_name] = {}
        for key, val in group_read.items():
          if key == 'attrs':
            attrs = dict(val.attrs)
            if 'order' in attrs:
              order = attrs.pop('order')
              ordered_dict = collections.OrderedDict()
              for ordered_key in order:
                ordered_dict.update({ordered_key: attrs[ordered_key]})
              attrs = ordered_dict
          else:
            for key, val in group_read.items():
              serialised[dist_name][key] = np.array(val)
              if serialised[dist_name][key].dtype in (np.dtype('S'), np.dtype('S21'), np.dtype('|S1')):
                serialised[dist_name][key] = set(val.shape[:-1])
        if attrs: 
          for k, v in attrs.items():
            if v == 'None':
              attrs[k] = None
          if order is not None:
            ordered_dict = collections.OrderedDict()
            for ordered_key in order:
              ordered_dict.update({ordered_key: serialised[dist_name][ordered_key]})
            for key, val in serialised[dist_name].items():
              if key not in ordered_dict:
                ordered_dict.update({key: val})
            serialised[dist_name] = ordered_dict
          serialised[dist_name].update({'attrs': attrs})
      else:
        aux_data[dist_name] = {}
        for key, val in group_read.items():
          aux_data[dist_name][key] = np.array(val)
  return serialised, aux_data

#-------------------------------------------------------------------------------
def write_dist(path, *args):
  return write_serialised(path, serialise(*args))

#-------------------------------------------------------------------------------
def read_dist(path):
  return deserialise(read_serialised(path)[0])

#-------------------------------------------------------------------------------



"""
A module that handles variable data types.
"""

#-------------------------------------------------------------------------------
import numpy as np
import sympy
import functools
import operator

#-------------------------------------------------------------------------------
VTYPES = {
          bool: {bool, np.dtype('bool')},
          int:  {int, np.dtype('int'), np.dtype('int32'), np.dtype('int64'),
                 np.uint, np.uint8, np.uint16, np.uint32, np.uint64},
          float: {float, np.dtype('float'),  np.dtype('float32'), np.dtype('float64')},
         }


OO = sympy.oo
OO_TO_NP = {OO: np.inf, -OO: -np.inf}

#-------------------------------------------------------------------------------
def eval_vtype(vtype):
  """ Evaluates a variable type from input vtype.

  :param vtype: input object

  :return: either bool, int, float if correctly detected. 
           Otherwise the dtype is returned.
  """
  if isinstance(vtype, set):
    vtype = list(vtype)
  if isinstance(vtype, (list, tuple)):
    if any([isinstance(_vtype, tuple) for _vtype in vtype]):
        vtype = np.concatenate([np.array(_vtype).reshape([1]) \
                                for _vtype in vtype]).dtype
    else:
      vtype = np.array(vtype)
  if hasattr(vtype, 'dtype'):
    vtype = vtype.dtype
  else:
    _vtype = vtype
    while _vtype != type:
      vtype = _vtype
      _vtype = type(vtype)
  if vtype in VTYPES.keys():
    return vtype
  for key, val in VTYPES.items():
    found = vtype in set(val)
    if found:
      vtype = key
      break
  return vtype

#-------------------------------------------------------------------------------
def isunitset(var, vtype=None):
  """ Returns boolean flag as to whether var is a set of one element.

  :param var: variable to assess.
  :param vtype: optional filter for variable type (e.g. int or float)
  """
  if vtype is None:
    vtype = list(VTYPES.keys())
  elif not isinstance(vtype, (tuple,list)):
    vtype = [vtype]
  vtypes = functools.reduce(operator.concat, [list(VTYPES[key]) for key in vtype])
  if isinstance(var, set):
    if len(var) == 1:
      element_type = type(list(var)[0])
      if element_type in vtypes:
        return True
  return False

#-------------------------------------------------------------------------------
def isunitsetint(var):
  """ Returns boolean flag as to whether var is a set of one integer element.
  Note: Usage depends on class:
  RVs, RJs, RFs: set(int) is a sample specification denoting number of samples:
                   positive values request samples using linear interpolation
                   negative values request samples using random generation.
  Dist: set(int): proxies as a 'value' for a variable as a set of size int.
  """
  return isunitset(var, int) 

#-------------------------------------------------------------------------------
def isunitsetfloat(var):
  """ Returns boolean flag as to whether var is a set of one float element.
  Usage requests a sampling of value from a ICDF for then given P.
  """
  return isunitset(var, float)

#-------------------------------------------------------------------------------
def isscalar(var):
  """ If var is a dimensionless numpy array of size 1 or if the var is a Sympy 
  Float or Integer, this function returns True, otherwise it returns 
  np.isscalar(var).
  """
  if isinstance(var, np.ndarray):
    if var.ndim == 0 and var.size == 1:
      return True
  if isinstance(var, (sympy.Float, sympy.Integer)):
    return True
  return np.isscalar(var)

#-------------------------------------------------------------------------------
def issingleton(var):
  """ If isunitset(var) is True, this function returns True, 
  otherwise isscalar(var) is returned.
  """
  # Here we define singleton as a unit set or scalar
  if isunitset(var):
    return True
  return isscalar(var)

#-------------------------------------------------------------------------------
def isunitary(var):
  """ If isscalar(var) is True, this function returns True, 
  otherwise if var is numpy array of size 1, this function returns True, 
  False is returned.
  """
  if isscalar(var):
    return True
  if isinstance(var, np.ndarray):
    if var.size == 1:
      return True
  return False

#-------------------------------------------------------------------------------
def isfinite(var):
  if isinstance(var, np.ndarray) or np.isscalar(var):
    return np.isfinite(var)
  if isinstance(var, (tuple, list)):
    isfin = [isfinite(subvar) for subvar in var]
    if isinstance(var, list):
      return isfin
    return tuple(isfin)
  if var == OO or var == -OO:
    return False
  return True

#-------------------------------------------------------------------------------
def isdimensionless(var):
  """ If var is {0}, then True is returned, otherwise np.isscalar(var) is
  returned.
  """
  if isinstance(var, set):
    if len(var) == 1:
      var = list(var)[0]
      if var == 0:
        return True
      return False
  return np.isscalar(var)

#-------------------------------------------------------------------------------
def revtype(var, vtype=None):
  """ Converts var into Numpy array of variable type vtype """
  if vtype is None:
    return var
  vtype = eval_vtype(vtype)
  vartype = eval_vtype(var)
  if vtype == vartype:
    return var
  if isscalar(var):
    return vtype(var)
  return np.array(var, dtype=vtype)

#-------------------------------------------------------------------------------
def uniform(v_0=0, 
            v_1=1, 
            n=None,
            ex_0=False,
            ex_1=False):
  r""" Samples $n$ points in the range $v_0$ to $v_1$ with optional exclusion of
  respective limits according to boolean flags values (default False) given by
  ex_0 and ex_1. The sampled values are uniformly obtained in the range 
  according to $n$:

  If $n<0$, then $-n$ points are sampled randomly (ignoring limit exclusions).
  If $n=0$, then a single scalar value is sampled randomly.
  If $n>0$, then $n$ points are linearly sampled according to limit exclusions.
  """

  # Zero or negative denote random uniform
  if not n:
    return np.random.uniform(v_0, v_1)
  if n < 0:
    return np.random.uniform(v_0, v_1, size=-n)

  # Using linspace may require slicing
  if not ex_0 and not ex_1:
    if n == 1:
      return np.linspace(v_0, v_1, 3)[1:-1]
    else:
      return np.linspace(v_0, v_1, n)
  if ex_0 and ex_1:
    return np.linspace(v_0, v_1, n+2)[1:-1]
  if ex_0 and not ex_1:
    return np.linspace(v_0, v_1, n+1)[1:]
  if not ex_0 and ex_1:
    return np.linspace(v_0, v_1, n+1)[:-1]

  # We shouldn't reach here
  raise ValueError("Unrecognised exclusions {} and {}".format(ex_0, ex_1))


#-------------------------------------------------------------------------------



"""
A probability class supporting probability distributions without specification 
of a variable set.
"""

#-------------------------------------------------------------------------------
import collections
import functools
import numpy as np
import scipy.stats
import sympy
from probayes.icon import isiconic
from probayes.pscales import eval_pscale, rescale, iscomplex
from probayes.vtypes import isscalar
from probayes.expression import Expression
from probayes.sympy_prob import is_sympy_stats_dist, SympyProb, sympy_sfun

#-------------------------------------------------------------------------------
SCIPY_STATS_CONT = {scipy.stats.rv_continuous}
SCIPY_STATS_DISC = {scipy.stats.rv_discrete}
SCIPY_STATS_MVAR = {scipy.stats._multivariate.multi_rv_generic}
SCIPY_STATS_DIST = SCIPY_STATS_MVAR.union(
                       SCIPY_STATS_CONT.union(SCIPY_STATS_DISC).\
                           union(SCIPY_STATS_MVAR))
SCIPY_DIST_METHODS = ['pdf', 'logpdf', 'pmf', 'logpmf', 'cdf', 'logcdf', 'ppf', 
                      'rvs', 'sf', 'logsf', 'isf', 'moment', 'stats', 'expect', 
                      'entropy', 'fit', 'median', 'mean', 'var', 'std', 'interval']

#-------------------------------------------------------------------------------
def is_scipy_stats_cont(arg, scipy_stats_cont=SCIPY_STATS_CONT):
  """ Returns if arg belongs to scipy.stats.continuous """
  return isinstance(arg, tuple(scipy_stats_cont))

def is_scipy_stats_disc(arg, scipy_stats_disc=SCIPY_STATS_DISC):
  """ Returns if arg belongs to scipy.stats.continuous """
  return isinstance(arg, tuple(scipy_stats_disc))

def is_scipy_stats_mvar(arg, scipy_stats_mvar=SCIPY_STATS_MVAR):
  """ Returns if arg belongs to scipy.stats._multivariate.multi_rv_generic """
  return isinstance(arg, tuple(scipy_stats_mvar))

def is_scipy_stats_dist(arg, scipy_stats_dist=SCIPY_STATS_DIST):
  """ Returns if arg belongs to scipy.stats.continuous or discrete """
  return isinstance(arg, tuple(scipy_stats_dist))

#-------------------------------------------------------------------------------
class Prob (Expression, SympyProb): 
  """ A probability is quantification of degrees of belief concerning outcomes.
  It is therefore an expression that can be defined with respect to no, one,
  or more variable.

  :example:
  >>> import scipy.stats
  >>> import probayes as pb
  >>> normprob = pb.Prob(scipy.stats.norm)
  >>> print(normprob(0.))
  0.3989422804014327
  >>> normlogp = pb.Prob(scipy.stats.norm, pscale='log')
  >>> print(normlogp(0.))
  -0.9189385332046727

  Use of Sympy's probabilistic expressions can be set w.r.t. a symbol
  :example:
  >>> import sympy; import sympy.stats
  >>> import probayes as pb
  >>> x = sympy.Symbol('x')
  >>> normprob = pb.Prob(sympy.stats.Normal(x, 0, 1))
  """

  # Protected
  _prob = None       # Probability distribution function
  _pscale = None     # Probability type (can be a scipy.stats.dist object)
  _logp = None       # Boolean flag to denote log probability from pscale
  _pfun = None       # 2-length tuple of cdf/icdf
  _sfun = None       # Random-variate sampling function

  # Private
  __isscipy = None   # Boolean flag of whether expression is a scipy stats object
  __issympy = None   # Boolean flag of whether expression is a sympy stats object
  __issmvar = None   # Boolean flag of whether expression is a scipy mvar object

#-------------------------------------------------------------------------------
  def __init__(self, prob=None, *args, **kwds):
    """ Initialises the probability and pscale (see set_prob). """
    self._pscale = None
    self.set_prob(prob, *args, **kwds)

#-------------------------------------------------------------------------------
  @property
  def prob(self):
    """ Returns probability value/object depending on set_prob(prob) argument:

    - if prob was an array, then its NumPy ndarray representation is returned.
    - if prob was a scalar, then its iconic Sympy representation is returned
    - if prob was a callable, then the corresponding function is returned.
    - if prob was a Sympy distribution/expression its expression is returned.
    - if prob was a Scipy distribution (e.g. scipy.stats.multivariate_normal), 
      then a partial call to its correponding probability function is returned.
    """
    return self._prob

  @property
  def isscipy(self):
    return self.__isscipy

  @property
  def issympy(self):
    return self.__issympy

  @property
  def issmvar(self):
    return self.__issmvar

  @property
  def pr(self):
    assert self.__issympy or self._isiconic, \
        "prob.pr only accessible for SymPy expressions"
    if self.__issympy:
      if self._logp:
        return self._exprs['logp']
      else:
        return self._exprs['prob']
    return self

  def set_prob(self, prob=None, *args, **kwds):
    """ Sets the probability and pscale with optional arguments and keywords.

    :param prob: may be a scalar, array, or callable function.
    :param *args: optional arguments to pass if prob is callable.
    :param **kwds: optional keywords to pass if prob is callable.

    'pscale' is a reserved keyword. See set_pscale() for explanation of how 
    pscale is used.
    """
    pscale = None if 'pscale' not in kwds else kwds.pop('pscale')
    self.pscale = pscale or self._pscale
    self.__isscipy = is_scipy_stats_dist(prob)
    self.__issympy = is_sympy_stats_dist(prob)
    self.__issmvar = is_scipy_stats_mvar(prob)

    # Probabilities can be defined as regular expressions, but iconise scalars
    if not self.__isscipy and not self.__issympy:
      prob_scalar = isscalar(prob)
      self._prob = prob if not prob_scalar else sympy.Float(prob)
      self.set_expr(self._prob, *args, **kwds)

      # Set probability callers to expression
      if not prob_scalar:
        self._prob = self._expr
      return # Expression() takes care of all partials

    # Scipy/SymPy expressions (self._expr set by _set_partials() later)
    if self.__issmvar or self.__issympy:
      self._expr = prob
      self._ismulti = True
      self._callable = True
      self._islambda = False
      self._isscalar = False
    else:
      self.set_expr(prob, *args, **kwds)
    self._args = tuple(args)
    self._kwds = dict(kwds)
    self._prob = self._expr
    if 'order' in self._kwds:
      self.set_order(self._kwds.pop('order'))
    if 'delta' in self._kwds:
      self.set_delta(self._kwds.pop('delta'))
    self._set_partials()

    # Scipy dist - set pfun and sfun calls - self._prob is updated but not used
    if self.__isscipy:
      self._ismulti = True
      self._prob = self._partials['logp'] if self._logp \
                   else self._partials['prob']
      if 'cdf' in self._keys and 'ppf' in self._keys:
        self.set_pfun((self._expr.cdf, self._expr.ppf), *self._args, **self._kwds)
      if 'rvs' in self._keys and hasattr(self._expr, 'rvs'):
        self.set_sfun(self._expr.rvs, *self._args, **self._kwds)

    # Sympy dist - set pfun and sfun calls - self._prob is updated but not used
    elif self.__issympy:
      self._prob = self._partials['logp'].expr if self._logp \
                   else self._partials['prob'].expr
      if 'cdf' in self._keys and 'icdf' in self._keys:
        self.set_pfun((self._partials['cdf'], self._partials['icdf']))
      if 'sfun' in self._keys:
        self._partials.update({'sfun': 
                               functools.partial(sympy_sfun, self._distr, dtype=float)})
      self.set_sfun(sympy_sfun, self._expr)

#-------------------------------------------------------------------------------
  def _set_partials(self):
    """ Overload Expression._set_partials() for Scipy and Sympy expressions """

    # Non-Scipy/Sympy expressions
    self._partials = collections.OrderedDict()
    if not self.__isscipy and not self.__issympy:
      super()._set_partials()
    
    # Extract SciPy object member functions
    elif self.__isscipy:

      # Set remaining scipy-specific flag
      self._isiconic = False

      # Instantiate multivariate scipy objects with pre-specified arguments
      if self.__issmvar and is_scipy_stats_mvar(self._expr):
        self._expr = self._expr(*self._args,  **self._kwds)
        self._args, self._kwds = (), {} # No longer used

      # Iterate available methods
      for method in SCIPY_DIST_METHODS:
        if hasattr(self._prob, method):
          call = functools.partial(Expression._partial_call, self, 
                                   getattr(self._expr, method),
                                   *self._args, **self._kwds)
          self._partials.update({method: call})

      # Provide a common interface for PDF/PMF and LOGPDF/LOGPMF
      if 'pdf' in self._partials.keys():
        self._partials.update({'prob': self._partials['pdf']})
        self._partials.update({'logp': self._partials['logpdf']})
      elif 'pmf' in self._partials.keys():
        self._partials.update({'prob': self._partials['pmf']})
        self._partials.update({'logp': self._partials['logpmf']})

    # Extract prob for sympy objects
    elif self.__issympy:
      self.distr = self._expr # This invokes the setter to define self._exprs
      self._partials.update(self._exprs)

    self._keys = list(self._partials.keys())

#-------------------------------------------------------------------------------
  @property
  def pscale(self):
    return self._pscale

  @property
  def logp(self):
    return self._logp

  @pscale.setter
  def pscale(self, pscale=None):
    """ Sets the probability scaling constant used for probabilities.

    :param pscale: can be None, a real number, or a complex number, or 'log'

       if pscale is None (default) the normalisation constant is set as 1.
       if pscale is real, this defines the normalisation constant.
       if pscale is complex, this defines the offset for log probabilities.
       if pscale is 'log', this denotes a logarithmic scale with an offset of 0.

    :return: pscale (either as a real or complex number)
    """
    self._pscale = eval_pscale(pscale)
    self._logp = iscomplex(self._pscale)
    return self._pscale

#-------------------------------------------------------------------------------
  @property
  def pfun(self):
    return self._pfun

  def set_pfun(self, pfun=None, *args, **kwds):
    """ Sets a two-length tuple of functions that should correspond to the
    (cumulative probability function, inverse cumulative function) with respect
    to the callable function set by set_prob(). It is necessary to set these
    functions if sampling variables non-randomly with non-flat distributions.

    :param pfun: two-length tuple of callable functions
    :param *args: arguments to pass to callable function
    :param **kwds: keywords to pass to callable function
    """
    self._pfun = pfun
    if self._pfun is None:
      return

    # Non-iconic ufun inputs must a two-length-tuple
    if not isiconic(self._pfun):
      message = "Input pfun must be a two-sized tuple of callable functions"
      assert isinstance(self._pfun, tuple), message
      assert len(self._pfun) == 2, message
      assert callable(self._pfun[0]), message
      assert callable(self._pfun[1]), message
    elif 'invertible' not in kwds:
      kwds.update({'invertible': True})
    self._pfun = Expression(self._pfun, *args, **kwds)

#-------------------------------------------------------------------------------
  @property
  def sfun(self):
    return self._sfun

  def set_sfun(self, sfun=None, *args, **kwds):
    """ Sets the random variate sampling function with respect to the callable 
    function set by set_prob(). It is necessary to set this functions if 
    sampling variables within infinite limits.

    :param sfun: a callable function or a two-tuple pair of functions
    :param *args: arguments to pass to callable function
    :param **kwds: keywords to pass to callable function

    If two functions are entered, the first is assumed for contiguous 
    unrandomised sampling, and the second for randomised.
    """
    self._sfun = sfun
    if self._sfun is None:
      return

    # Non-iconic sfun inputs must a two-length-tuple
    if not isiconic(self._sfun):
      message = "Input sfun must be a single or pair of callable functions"
      if isinstance(self._sfun, tuple):
        assert len(self._sfun) == 2, message
        assert callable(self._sfun[0]), message
        assert callable(self._sfun[1]), message
      else:
        assert callable(self._sfun), message
    self._sfun = Expression(self._sfun, *args, **kwds)

#-------------------------------------------------------------------------------
  def rescale(self, probs, **kwds):
    """ Returns a rescaling of probs from current pscale to the values according
    to the keyword pscale=new_pscale. """
    if 'pscale' not in kwds:
      return probs
    return rescale(probs, self._pscale, kwds['pscale'])

#-------------------------------------------------------------------------------
  def eval_prob(self, *args, **kwds):
    """ Evaluates the probability inputting optional args for callable cases

    :param *args: optional arguments for callable probability objects.
    :param **kwds: optional arguments to include pscale for rescaling.

    :return: evaluated probabilities
    """
    # Strip pscale if pscale
    kwds = dict(kwds)
    pscale = None if 'pscale' not in kwds else kwds.pop('pscale')

    # Callable and non-callable evaluations
    if self.callable:

      # Scipy
      if self.__isscipy:
        prob = self._partials['logp'] if self._logp else self._partials['prob']
        if self.issmvar: # for mvar, convert dictionaries to arrays
          vals = list(args[0].values())
          if len(vals) == 1:
            vals = np.array(vals)
          else:
            vals = vals[::-1]
            if len(vals) > 2:
              vals = vals[1:] + [vals[0]]
            vals = np.stack(np.meshgrid(*tuple(vals)), axis=-1)
          prob = prob(vals)
        else:
          prob = prob(*args)

      # Sympy distributions are looked after by SympyProb
      elif self.__issympy:
        prob = self._partials['logp'] if self._logp else self._partials['prob']
        prob = prob(*args, **kwds)

      # Expression callables handled by expression partials interface
      else:
        prob = self._partials[None](*args, **kwds)

    # Non-callables must not have any arguments
    else:
      assert not len(args) and not len(kwds), \
          "Cannot evaluate from values from an uncallable probability function"
      prob = self._prob

    # Optionally rescale
    if pscale:
      return self.rescale(prob, kwds['pscale'])
    return prob

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
    """ See eval_prob() """
    return Prob.eval_prob(self, *args, **kwds)

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
    """ Returns partial function for arg. For Sympy statistical distributions,
    additional returns are supported:
    If set(), then the original distribution is returned.
    If [], then the distribution probability class object is returned.
    If {}, the full expression dictionary is returned.
    If (), the logp or prob expression (depending on pscale) is returned
    If '', then the keys of the partials dictionary are returned.
    Note arg cannot be a slice object.
    """
    if isinstance(arg, slice):
      raise TypeError("Argument input cannot be slice or tuple type")
    if isinstance(arg, (list, dict, set, tuple)):
      assert self.__issympy, \
          "Input argument of type {} supported only Sympy distributions".type(arg)
      assert not len(arg), \
          "Input argument of type {} must be empty".type(arg)
      if isinstance(arg, set):
        return self._distr
      if isinstance(arg, list):
        return self._probj
      if isinstance(arg, dict):
        return self._exprs
      if isinstance(arg, tuple):
        if self._logp:
          return self._exprs['logp'].expr
        return self._exprs['prob'].expr
    if isinstance(arg, str) and not len(arg):
      return list(self._exprs.keys())
    return Expression.__getitem__(self, arg)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()

#-------------------------------------------------------------------------------



""" Random variable module """

#-------------------------------------------------------------------------------
import collections
import warnings
import numpy as np
import sympy
from probayes.variable import Variable, DEFAULT_VNAME
from probayes.prob import Prob
from probayes.vtypes import uniform, VTYPES, isscalar, \
                        isunitset, isunitsetint, isunitsetfloat, issingleton
from probayes.pscales import rescale
from probayes.expression import Expression
from probayes.sympy_prob import bernoulli_prob, bernoulli_sfun
from probayes.rv_utils import uniform_prob, matrix_cond_sample, \
                          lookup_square_matrix
from probayes.distribution import Distribution
from probayes.pd import PD

"""
A random variable is a triple (x, A_x, P_x) defined for an outcome x for every 
possible realisation defined over the alphabet set A_x with probabilities P_x.
It therefore requires a name for x (id), a variable alphabet set (vset), and its 
associated probability distribution function (prob).
"""

#-------------------------------------------------------------------------------
class RV (Variable, Prob):
  """ A random variable is a variable with a defined probability function.
  It therefore inherits from classes Variable and and Prob. Each instance therefore 
  requires a name, a variable set, and probability function. Additionally RV
  supports transitional probabilities the cdf/icdf equivalents specified using
  RV.set_tran() and RV.set_tfun() equivalents, and accessed using RV.step().

  :example:
  >>> import numpy as np
  >>> import probayes as pb
  >>> var = pb.RV('var', vtype=bool)
  >>> var.set_tran(np.array([0.2, 0.8, 0.3, 0.7]).reshape(2,2,))
  >>> step = var.step()
  >>> print(step.prob)
  [[0.2 0.8]
   [0.3 0.7]]
  """

  # Protected
  _tran = None          # Transitional prob - can be a matrix
  _tfun = None          # Like pfun for transitional conditionals
  _is_stochastic = True # Flag of whether stochastic

  # Private
  __def_prob = None   # Flag to denote prob is defaulted
  __sym_tran = None   # Flag to denote symmetric transitional
  __prime_key = None  # Modified-string to denote prime key for variable

#-------------------------------------------------------------------------------
  def __init__(self, name, 
                     vset=None, 
                     vtype=None,
                     prob=None,
                     *args,
                     **kwds):
    """ Initialises a random variable combining Variable and Prob initialisation
    except invertible monotonic must be specified separately using set_ufun().

    :param name: Name of the domain - string as valid identifier.
    :param vset: variable set over which domain defined (see Variable.vset).
    :param vtype: variable type (bool, int, or float).
    :param prob : uncallable or callable probability expression (see set_prob).
    :param *args: optional arguments to pass if prob is callable.
    :param **kwds: optional keywords to pass if prob is callable.
    """

    Variable.__init__(self, name, vtype, vset)
    Prob.__init__(self, prob, *args, **kwds)
    self._tran, self._tfun = None, None

#-------------------------------------------------------------------------------
  @property
  def name(self):
    return self._name

  @name.setter
  def name(self, name=DEFAULT_VNAME):
    self._name = name
    self._Delta = collections.namedtuple('ð', [self._name])
    self.__prime_key = self._name + "'"

#-------------------------------------------------------------------------------
  @property
  def vset(self):
    return self._vset

  @Variable.vset.setter
  def vset(self, vset=None):
    Variable.vset.fset(self, vset)
    if self.__def_prob:
      self._prob = None
      self._default_prob()

#-------------------------------------------------------------------------------
  @property
  def vtype(self):
    return self._vtype

  @Variable.vtype.setter
  def vtype(self, vtype=None):
    Variable.vtype.fset(self, vtype)
    if self.__def_prob:
      self._prob = None
      self._default_prob()

#-------------------------------------------------------------------------------
  @property
  def prob(self):
    return self._prob

  @property
  def def_prob(self):
    return self.__def_prob

  def set_prob(self, prob=None, *args, **kwds):
    """ Sets the probability and pscale with optional arguments and keywords.

    :param prob: may be a scalar, array, or callable function.
    :param *args: optional arguments to pass if prob is callable.
    :param **kwds: optional keywords to pass if prob is callable.

    'pscale' is a reserved keyword. See Prob.pscale for explanation of how 
    pscale is used.
    """
    super().set_prob(prob, *args, **kwds)
    if self._prob is None:
      self._default_prob()
    else:
      self.__def_prob = False

    # Check uncallable probabilities commensurate with self._vset
    if self._vset is not None and \
        not self.callable and not self.isscalar:
      assert len(self._prob) == len(self._vset), \
          "Probability of length {} incommensurate with Vset of length {}".format(
              len(self._prob), len(self._vset))

    # Reassign non-0.5 Bernoulli probabilities to a formal distribution
    if self._vtype in VTYPES[bool] and self.isscalar:
      prob = float(self._prob)
      if not np.isclose(prob, 0.5):
        super().set_prob(bernoulli_prob(self.icon, bias=prob))
        super().set_sfun(bernoulli_sfun, bias=prob)

#-------------------------------------------------------------------------------
  def _default_prob(self):
    """ Defaults unspecified probabilities to uniform over self._vset.
    This will not override a previously specified non-default probability.
    """

    if self._prob is not None and not self.__def_prob:
      return
    prob = rescale(self._nlhv, 'log', self._pscale)
    if self._ufun is not None and self.no_ucov:
      prob = prob + sympy.log(self._ufun.derinv.expr) if self._logp else \
             prob * self._ufun.derinv.expr
    super().set_prob(prob)
    self.set_tran(prob)
    self.__def_prob = True

#-------------------------------------------------------------------------------
  def set_pfun(self, pfun=None, *args, **kwds):
    """ Sets a two-length tuple of functions that should correspond to the
    (cumulative probability function, inverse cumulative function) with respect
    to the callable function set by set_prob(). It is necessary to set these
    functions if sampling variables with non-flat distributions.

    :param pfun: two-length tuple of callable functions
    :param *args: arguments to pass to callable functions
    :param **kwds: keywords to pass to callable functions
    """
    super().set_pfun(pfun, *args, **kwds)
    if self._ufun is None or self._pfun is None:
      return
    assert self._ufun is None, \
      "Cannot assign non-uniform distribution alongside " + \
      "values transformation functions"

#-------------------------------------------------------------------------------
  def set_ufun(self, ufun=None, *args, **kwds):
    """ Sets a monotonic invertible tranformation for the domain as a tuple of
    two functions in the form (transforming_function, inverse_function) 
    operating on the first argument with optional further args and kwds.

    :param ufun: two-length tuple of monotonic functions.
    :param *args: args to pass to ufun functions.
    :param **kwds: kwds to pass to ufun functions.

    Support for this transformation is only valid for float-type vtypes.
    """
    super().set_ufun(ufun, *args, **kwds)
    if self._ufun is None:
      return

    # Recalibrate defaulted probabilities for floating point vtypes
    if self._vtype in VTYPES[float]:
      self._default_prob()

    # Assert pfun is unspecified
    assert self._pfun is None, \
      "Cannot assign univariate function alongside CDF/ICDF specification"

#-------------------------------------------------------------------------------
  @property
  def tran(self):
    """ Returns the transitional probability expression if specified """
    return self._tran

  def set_tran(self, tran=None, *args, **kwds):
    """ Sets a transitional function as a conditional probability. This can
    be specified numerically or one or two callable functions.

    :param tran: conditional scalar, array, or callable function (see below).
    :param *args: args to pass to tran functions.
    :param **kwds: kwds to pass to tran functions.

    If tran is a scalar, array, or callable function, then the transitional
    conditionality is treated as symmetrical. If tran is a two-length tuple,
    then assymetry is assumed in the form: (p[var'|var], p[var|var']).

    If intending to sample from a transitional conditional probability density
    function, the corresponding (CDF, ICDF) must be set using set_tfun().
    """
    self._tran = tran
    self.__sym_tran = None
    if self._tran is None:
      return
    self._tran = Expression(self._tran, *args, **kwds)
    self.__sym_tran = not self._tran.ismulti
    if self._tran.callable or self._tran.isscalar:
      return
    assert self._vtype not in VTYPES[float],\
      "Scalar or callable transitional required for floating point data types"
    tran = self._tran() if self.__sym_tran else self._tran[0]()
    message = "Transition matrix must a square 2D Numpy array " + \
              "covering variable set of size {}".format(len(self._vset))
    assert isinstance(tran, np.ndarray), message
    assert tran.ndim == 2, message
    assert np.all(np.array(tran.shape) == len(self._vset)), message
    self.__sym_tran = np.allclose(tran, tran.T)

#-------------------------------------------------------------------------------
  @property
  def tfun(self):
    """ Returns the transitional CDF/ICDF expression if specified """
    return self._tfun

  def set_tfun(self, tfun=None, *args, **kwds):
    """ Sets a two-length tuple of functions that should correspond to the
    (cumulative probability function, inverse cumulative function) with respect
    to the callable function set by set_tran(). It is necessary to set these
    functions if conditionally sampling variables with continuous distributions.

    :param tfun: two-length tuple of callable functions
    :param *args: arguments to pass to tfun functions
    :param **kwds: keywords to pass to tfun functions
    """
    self._tfun = tfun if tfun is None else Expression(tfun, *args, **kwds)
    if self._tfun is None:
      return
    assert self._tfun.ismulti, "Tuple of two functions required"

#-------------------------------------------------------------------------------
  def evaluate(self, values, use_pfun=True):
    """ Evaluates value(s) belonging to the domain of the variable.

    :param values: None, set of a single integer, array, or scalar.
    :param use_pfun: boolean flag to make use of pfun if previously set.

    :return: a NumPy array of the values (see Variable.evaluate()):
    """
    use_pfun = use_pfun and self._pfun is not None and isunitsetint(values)

    # If not using pfun while random sampling, we use sfun for booleans
    if not use_pfun:
      if self._vtype in VTYPES[bool] and hasattr(self._sfun, 'expr'):
        if isunitsetint(values) and self._sfun.expr == bernoulli_sfun:
          number = list(values)[0]
          if not number or number < 0:
            number = number if not number else -number
            return Distribution(self._name, 
                                {self.name: self._sfun[None](number)})
      return super().evaluate(values)

    # Evaluate values from inverse cdf bounded within cdf limits
    number = list(values)[0]
    assert self.isfinite, \
        "Cannot evaluate {} values for bounds: {}".format(values, self._ulims)
    lims = self.pfun[0](self._ulims)
    values = uniform(
                     lims[0], lims[1], number, 
                     isinstance(self._vset[0], tuple),
                     isinstance(self._vset[1], tuple)
                    )
    return Distribution(self._name, {self.name: self.pfun[1](values)})

#-------------------------------------------------------------------------------
  def eval_prob(self, values=None):
    """ Evaluates the probability inputting optional args for callable cases

    :param values: values of the variable used for evaluating probabilities.
    :param *args: optional arguments for callable probability objects.
    :param **kwds: optional arguments to include pscale for rescaling.

    :return: evaluated probabilities
    """
    values = values[self.name] if isinstance(values, dict) else values
    if not self.isscalar:
      return super().eval_prob(values)
    return uniform_prob(values, 
                        prob=float(self._prob), 
                        inside=self._inside,
                        pscale=self._pscale)

#-------------------------------------------------------------------------------
  def eval_dist_name(self, values, suffix=None):
    """ Evaluates a distribution name for a probability distribution based on
    the values set in the first input argument with an optional suffix. """
    name = self._name if not suffix else self._name + suffix
    if values is None:
      dist_str = name
    elif np.isscalar(values):
      dist_str = "{}={}".format(name, values)
    else:
      dist_str = name + "=[]"
    return dist_str

#-------------------------------------------------------------------------------
  def eval_step(self, pred_vals, succ_vals, reverse=False):
    """ Evaluates a successive values from previous values with an optional
    direction reversal flag, outputting a three-length tuple that includes the
    successive values in the first argument.

    :param pred_vals: predecessor values (NumPy array).
    :param succ_vals: succecessor values (see step()).
    :param reverse: boolean flag (default False) to reverse direction.

    :return vals: a dictionary including both predecessor and successor values.
    :return dims: a dictionary with dimension indices for the values in vals.
    :return kwargs: a dictionary that includes optional keywords for eval_tran()
    """

    if succ_vals is None:
      assert self._tran is not None, "No transitional function specified"
    if isinstance(pred_vals, dict):
      pred_vals = pred_vals[self.name]
    kwargs = dict() # to pass over to eval_tran()
    if succ_vals is None:
      if self._delta is None:
        succ_vals = {0} if isscalar(pred_vals) else pred_vals
      else:
        delta = self.eval_delta()
        succ_vals = self.apply_delta(pred_vals, delta)

    #---------------------------------------------------------------------------
    def _reshape_vals(pred, succ):
      dims = {}
      ndim = 0

      # Now reshape the values according to succ > prev dimensionality
      if issingleton(succ):
        dims.update({self._name+"'": None})
      else:
        dims.update({self._name+"'": ndim})
        ndim += 1
      if issingleton(pred):
        dims.update({self._name: None})
      else:
        dims.update({self._name: ndim})
        ndim += 1

      if ndim == 2: # pred_vals distributed along inner dimension:
        pred = pred.reshape([1, pred.size])
        succ = succ.reshape([succ.size, 1])
      return pred, succ, dims

    #---------------------------------------------------------------------------
    # Scalar treatment is the most trivial and ignores reverse
    if self._tran is None or self._tran.isscalar:
      if isunitsetint(succ_vals):
        succ_vals = self.evaluate(succ_vals, use_pfun=False)[self._name]
      elif isunitsetfloat(succ_vals):
        assert self._vtype in VTYPES[float], \
            "Inverse CDF sampling for scalar probabilities unavailable for " + \
            "{} data type".format(self._vtype)
        cdf_val = list(succ_vals)[0]
        lo, hi = min(self._limits), max(self._limits)
        succ_val = lo*(1.-cdf_val) + hi*cdf_val
        if self._ufun is not None:
          succ_val = self.ufun[-1](succ_val)

      prob = self._tran() if self._tran is not None else None
      pred_vals, succ_vals, dims = _reshape_vals(pred_vals, succ_vals)
                  
    # Handle discrete non-callables
    elif not self._tran.callable:
      if reverse and not self._tran.ismulti and not self.__sym_tran:
        warnings.warn("Reverse direction called from asymmetric transitional")
      prob = self._tran() if not self._tran.ismulti else \
             self._tran[int(reverse)]()
      if isunitset(succ_vals):
        succ_vals, pred_idx, succ_idx = matrix_cond_sample(pred_vals, 
                                                           succ_vals, 
                                                           prob=prob, 
                                                           vset=self._vset) 
        kwargs.update({'pred_idx': pred_idx, 'succ_idx': succ_idx})
      pred_vals, succ_vals, dims = _reshape_vals(pred_vals, succ_vals)

    # That just leaves callables
    else:
      kwds = {self._name: pred_vals}
      if isunitset(succ_vals):
        assert self._tfun is not None, \
            "Conditional sampling requires setting CDF and ICDF " + \
            "conditional functions using rv.set.tfun()"
        assert isscalar(pred_vals), \
            "Successor sampling only possible with scalar predecessors"
        succ_vals = list(succ_vals)[0]
        if type(succ_vals) in VTYPES[int] or type(succ_vals) in VTYPES[np.uint]:
          lo, hi = min(self._ulims), max(self._ulims)
          kwds.update({self._name+"'": np.array([lo, hi], dtype=float)})
          lohi = self._tfun[0](**kwds)
          lo, hi = float(min(lohi)), float(max(lohi))
          succ_vals = uniform(lo, hi, succ_vals,
                              isinstance(self._vset[0], tuple),
                              isinstance(self._vset[1], tuple))
        else:
          succ_vals = np.atleast_1d(succ_vals)
        kwds.update({self._name: pred_vals,
                     self._name+"'": succ_vals})
        succ_vals = self._tfun[1](**kwds)
      elif not isscalar(succ_vals):
        succ_vals = np.atleast_1d(succ_vals)
      pred_vals, succ_vals, dims = _reshape_vals(pred_vals, succ_vals)

    vals = collections.OrderedDict({self._name+"'": succ_vals,
                                    self._name: pred_vals})
    kwargs.update({'reverse': reverse})
    return vals, dims, kwargs

#-------------------------------------------------------------------------------
  def eval_tran(self, vals, **kwargs):
    """ Evaluates the transitional conditional probability for the dictionary 
    arguments in vals with optional keywords in **kwargs.
    """
    reverse = False if 'reverse' not in kwargs else kwargs['reverse']
    pred_vals, succ_vals = vals[self._name], vals[self._name+"'"]
    pred_idx = None if 'pred_idx' not in kwargs else kwargs['pred_idx'] 
    succ_idx = None if 'succ_idx' not in kwargs else kwargs['succ_idx'] 
    cond = None

    # No transitional means no conditional
    if self._tran is None:
      pass

    # Scalar treatment is the most trivial and ignores reverse
    elif self._tran.isscalar:
      prob = None if self._tran is None else self._tran()
      cond = uniform_prob(pred_vals,
                          succ_vals, 
                          prob=float(prob), 
                          inside=self._inside) 
                  

    # Handle discrete non-callables
    elif not self._tran.callable:
      prob = self._tran() if not self._tran.ismulti else \
             self._tran[int(reverse)]()
      cond = lookup_square_matrix(pred_vals,
                                  succ_vals, 
                                  sq_matrix=prob, 
                                  vset=self._vset,
                                  col_idx=pred_idx,
                                  row_idx=succ_idx) 


    # That just leaves callables
    else:
      prob = self._tran if not self._tran.ismulti else \
             self._tran[int(reverse)]
      kwds = {self._name: pred_vals,
              self._name+"'": succ_vals}
      cond = prob(**kwds)

    return cond

#-------------------------------------------------------------------------------
  def __call__(self, values=None):
    """ Return a probability distribution for the quantities in values. """
    dist_name = self.eval_dist_name(values)
    vals = self.evaluate(values)
    prob = self.eval_prob(vals)
    dims = {self._name: None} if isscalar(vals[self.name]) else {self._name: 0}
    return PD(dist_name, vals, dims=dims, prob=prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def step(self, *args, reverse=False):
    """ Returns a conditional probability distribution for quantities in args.

    :param *args: predecessor, successor values to evaluate conditionals.
    :param reverse: Boolean flag to evaluate conditional probability in reverse.

    :return a Dist instance of the conditional probability distribution
    """
    pred_vals, succ_vals = None, None 
    if len(args) == 1:
      if isinstance(args[0], (list, tuple)) and len(args[0]) == 2:
        pred_vals, succ_vals = args[0][0], args[0][1]
      else:
        pred_vals = args[0]
    elif len(args) == 2:
      pred_vals, succ_vals = args[0], args[1]
    dist_pred_name = self.eval_dist_name(pred_vals)
    dist_succ_name = None
    if pred_vals is None and succ_vals is None and \
        self._vtype not in VTYPES[float]:
      dist_succ_name = self.eval_dist_name(succ_vals, "'")
    pred_vals = self.evaluate(pred_vals)
    vals, dims, kwargs = self.eval_step(pred_vals, succ_vals, reverse=reverse)
    cond = self.eval_tran(vals, **kwargs)
    if dist_succ_name is None:
      dist_succ_name = self.eval_dist_name(vals[self.__prime_key], "'")
    dist_name = '|'.join([dist_succ_name, dist_pred_name])

    # TODO - distinguish between probabilistic and non-probabistic outputs
    if cond is None:
      cond = 1.
      for val in vals.values():
        if isinstance(val, np.ndarray):
          cond = cond * np.ones(val.shape, dtype=float)

    return PD(dist_name, vals, dims=dims, prob=cond, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def __and__(self, other):
    """ Combination operator between RV and another RV, RF, or SD. """
    from probayes.rf import RF
    from probayes.sd import SD
    if isinstance(other, SD):
      leafs = [self] + list(other.leafs.vars.values())
      stems = other.stems
      roots = other.roots
      args = RF(*tuple(leafs))
      if stems:
        args += list(stems.values())
      if roots:
        args += list(roots.values())
      return SD(*args)

    if isinstance(other, RF):
      rvs = [self] + list(other.vars.values())
      return RF(*tuple(rvs))

    if isinstance(other, RV):
      return RF(self, other)

    raise TypeError("Unrecognised post-operand type {}".format(type(other)))

#-------------------------------------------------------------------------------
  def __or__(self, other):
    """ Conditional operator between RV and another RV, RF, or SD. """
    from probayes.sd import SD
    return SD(self, other)

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
    """ If arg is a slice (i.e. RV[:]) then Variable icon is returned,
    otherwise the Prob.__getitem__ method is called.
    """
    if isinstance(arg, slice):
      return Variable.__getitem__(self, arg)
    return Prob.__getitem__(self, arg)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
def RVs(names, vtype=None, vset=None, prob=None, *args, **kwds):
  if isinstance(names, str):
    names = names.split(',')
  varlist = [RV(name, vtype, vset, prob, *args, **kwds) for name in names]
  return tuple(varlist)

#-------------------------------------------------------------------------------
